@@ -53,7 +53,7 @@ liberator.Bookmarks = function () //{{{
     function load()
     {
         // update our bookmark cache
-        bookmarks = []; // also clear our bookmark cache
+        bookmarks = [];
         keywords  = [];
 
         var folders = [bookmarksService.toolbarFolder, bookmarksService.bookmarksMenuFolder, bookmarksService.unfiledBookmarksFolder];
@@ -101,6 +101,7 @@ liberator.Bookmarks = function () //{{{
     liberator.options.add(["defsearch", "ds"],
         "Set the default search engine",
         "string", "google");
+
     liberator.options.add(["preload"],
         "Speed up first time history/bookmark completion",
         "boolean", true);
@@ -129,20 +130,42 @@ liberator.Bookmarks = function () //{{{
     ////////////////////// COMMANDS ////////////////////////////////////////////////
     /////////////////////////////////////////////////////////////////////////////{{{
 
+    liberator.commands.add(["ju[mps]"],
+        "Show jumplist",
+        function ()
+        {
+            var sh = getWebNavigation().sessionHistory;
+            var list = ":" + (liberator.util.escapeHTML(liberator.commandline.getCommand()) || "jumps") + "<br/>" + "<table>";
+            list += "<tr style=\"text-align: left; color: magenta; font-weight: bold;\"><th colspan=\"2\">jump</th><th>title</th><th>URI</th></tr>";
+            var num = -sh.index;
+
+            for (var i = 0; i < sh.count; i++)
+            {
+                var entry = sh.getEntryAtIndex(i, false);
+                var uri = entry.URI.spec;
+                var title = entry.title;
+                var indicator = i == sh.index? "<span style=\"color: blue;\">&gt;</span>": " ";
+                list += "<tr><td>" + indicator + "<td>" + Math.abs(num) + "</td><td style=\"width: 250px; max-width: 500px; overflow: hidden;\">" + title +
+                        "</td><td><a href=\"#\" class=\"hl-URL jump-list\">" + uri + "</a></td></tr>";
+                num++;
+            }
+
+            list += "</table>";
+
+            liberator.commandline.echo(list, liberator.commandline.HL_NORMAL, liberator.commandline.FORCE_MULTILINE);
+        },
+        { argCount: "0" });
+
     liberator.commands.add(["bma[rk]"],
         "Add a bookmark",
         function (args)
         {
-            var res = liberator.commands.parseArgs(args, this.args);
-            if (!res)
-                return;
-
-            var url = res.args.length == 0 ? liberator.buffer.URL : res.args[0];
-            var title = liberator.commands.getOption(res.opts, "-title", res.args.length == 0 ? liberator.buffer.title : null);
+            var url = args.arguments.length == 0 ? liberator.buffer.URL : args.arguments[0];
+            var title = args["-title"] || (args.arguments.length == 0 ? liberator.buffer.title : null);
             if (!title)
                 title = url;
-            var keyword = liberator.commands.getOption(res.opts, "-keyword", null);
-            var tags = liberator.commands.getOption(res.opts, "-tags", []);
+            var keyword = args["-keyword"] || null;
+            var tags =    args["-tags"] || [];
 
             if (liberator.bookmarks.add(false, title, url, keyword, tags))
             {
@@ -155,25 +178,21 @@ liberator.Bookmarks = function () //{{{
                 liberator.echoerr("Exxx: Could not add bookmark `" + title + "'", liberator.commandline.FORCE_SINGLELINE);
         },
         {
-            args: [[["-title", "-t"],    liberator.commands.OPTION_STRING],
-                   [["-tags", "-T"],     liberator.commands.OPTION_LIST],
-                   [["-keyword", "-k"],  liberator.commands.OPTION_STRING, function (arg) { return /\w/.test(arg); }]]
+            options: [[["-title", "-t"],    liberator.commands.OPTION_STRING],
+                      [["-tags", "-T"],     liberator.commands.OPTION_LIST],
+                      [["-keyword", "-k"],  liberator.commands.OPTION_STRING, function (arg) { return /\w/.test(arg); }]],
+            argCount: "?"
         });
 
     liberator.commands.add(["bmarks"],
         "List or open multiple bookmarks",
         function (args, special)
         {
-            var res = liberator.commands.parseArgs(args, this.args);
-            if (!res)
-                return;
-
-            var tags = liberator.commands.getOption(res.opts, "-tags", []);
-            liberator.bookmarks.list(res.args.join(" "), tags, special);
+            liberator.bookmarks.list(args.arguments.join(" "), args["-tags"] || [], special);
         },
         {
             completer: function (filter) { return [0, liberator.bookmarks.get(filter)]; },
-            args: [[["-tags", "-T"], liberator.commands.OPTION_LIST]]
+            options: [[["-tags", "-T"], liberator.commands.OPTION_LIST]]
         });
 
     liberator.commands.add(["delbm[arks]"],
@@ -239,7 +258,7 @@ liberator.Bookmarks = function () //{{{
             }
             catch (e)
             {
-                liberator.log(e);
+                liberator.log(e, 0);
                 return false;
             }
 
@@ -306,7 +325,7 @@ liberator.Bookmarks = function () //{{{
             }
             catch (e)
             {
-                liberator.log(e);
+                liberator.log(e, 0);
                 return i;
             }
 
@@ -325,7 +344,7 @@ liberator.Bookmarks = function () //{{{
         getSearchEngines: function ()
         {
             var searchEngines = [];
-            var firefoxEngines = searchService.getVisibleEngines({ });
+            var firefoxEngines = searchService.getVisibleEngines({});
             for (var i in firefoxEngines)
             {
                 var alias = firefoxEngines[i].alias;
@@ -364,53 +383,26 @@ liberator.Bookmarks = function () //{{{
             return keywords;
         },
 
-        // if @param engineName is null, it uses the default search engine
+        // full search string including engine name as first word in @param text
+        // if @param useDefSearch is true, it uses the default search engine
         // @returns the url for the search string
         //          if the search also requires a postData, [url, postData] is returned
-        getSearchURL: function (text, engineName)
+        getSearchURL: function (text, useDefsearch)
         {
             var url = null;
-            var postData = null;
-            if (!engineName)
-                engineName = liberator.options["defsearch"];
+            var aPostDataRef = {};
+            var searchString = (useDefsearch? liberator.options["defsearch"] + " " : "") + text;
 
             // we need to make sure our custom alias have been set, even if the user
             // did not :open <tab> once before
             this.getSearchEngines();
 
-            // first checks the search engines for a match
-            var engine = searchService.getEngineByAlias(engineName);
-            if (engine)
-            {
-                if (text)
-                {
-                    var submission = engine.getSubmission(text, null);
-                    url = submission.uri.spec;
-                    postData = submission.postData;
-                }
-                else
-                    url = engine.searchForm;
-            }
-            else // check for keyword urls
-            {
-                if (!keywords)
-                    load();
+            url = getShortcutOrURI(searchString, aPostDataRef);
+            if (url == searchString)
+                url = null;
 
-                for (var i in keywords)
-                {
-                    if (keywords[i][0] == engineName)
-                    {
-                        if (text == null)
-                            text = "";
-                        url = keywords[i][2].replace(/%s/g, encodeURIComponent(text));
-                        break;
-                    }
-                }
-            }
-
-            // if we came here, the engineName is neither a search engine or URL
-            if (postData)
-                return [url, postData];
+            if (aPostDataRef && aPostDataRef.value)
+                return [url, aPostDataRef.value];
             else
                 return url; // can be null
         },
@@ -421,8 +413,12 @@ liberator.Bookmarks = function () //{{{
             var items = this.get(filter, tags, false);
             if (items.length == 0)
             {
-                if (filter.length > 0 || tags.length > 0)
-                    liberator.echoerr("E283: No bookmarks matching \"" + filter + "\"");
+                if (filter.length > 0 && tags.length > 0)
+                    liberator.echoerr("E283: No bookmarks matching tags: \"" + tags + "\" and string: \"" + filter + "\"");
+                else if (filter.length > 0)
+                    liberator.echoerr("E283: No bookmarks matching string: \"" + filter + "\"");
+                else if (tags.length > 0)
+                    liberator.echoerr("E283: No bookmarks matching tags: \"" + tags + "\"");
                 else
                     liberator.echoerr("No bookmarks set");
 
@@ -487,7 +483,8 @@ liberator.History = function () //{{{
     const historyService = Components.classes["@mozilla.org/browser/nav-history-service;1"]
                            .getService(Components.interfaces.nsINavHistoryService);
 
-    var history = null;
+    var history = [];
+    var cachedHistory = []; // add pages here after loading the initial Places history
 
     if (liberator.options["preload"])
         setTimeout(function () { load(); }, 100);
@@ -509,12 +506,9 @@ liberator.History = function () //{{{
         for (var i = 0; i < rootNode.childCount; i++)
         {
             var node = rootNode.getChild(i);
+            // dump("History child " + node.itemId + ": " + node.title + " - " + node.type + "\n");
             if (node.type == node.RESULT_TYPE_URI) // just make sure it's a bookmark
-            {
-                history.push([node.uri, node.title]);
-            }
-            else
-                dump("History child " + node.itemId + ": " + node.title + " - " + node.type + "\n");
+                history.push([node.uri, node.title || "[No title]"]);
         }
 
         // close a container after using it!
@@ -556,7 +550,9 @@ liberator.History = function () //{{{
         function (args, special, count)
         {
             if (special)
+            {
                 liberator.history.goToStart();
+            }
             else
             {
                 if (args)
@@ -570,8 +566,12 @@ liberator.History = function () //{{{
                             return;
                         }
                     }
+                    liberator.echoerr("Exxx: URL not found in history");
+                }
+                else
+                {
+                    liberator.history.stepTo(count > 0 ? -1 * count : -1);
                 }
-                liberator.history.stepTo(count > 0 ? -1 * count : -1);
             }
         },
         {
@@ -596,7 +596,9 @@ liberator.History = function () //{{{
         function (args, special, count)
         {
             if (special)
+            {
                 liberator.history.goToEnd();
+            }
             else
             {
                 if (args)
@@ -610,8 +612,12 @@ liberator.History = function () //{{{
                             return;
                         }
                     }
+                    liberator.echoerr("Exxx: URL not found in history");
+                }
+                else
+                {
+                    liberator.history.stepTo(count > 0 ? count : 1);
                 }
-                liberator.history.stepTo(count > 0 ? count : 1);
             }
         },
         {
@@ -635,6 +641,7 @@ liberator.History = function () //{{{
         "Show recently visited URLs",
         function (args, special) { liberator.history.list(args, special); },
         { completer: function (filter) { return [0, liberator.history.get(filter)]; } });
+
     /////////////////////////////////////////////////////////////////////////////}}}
     ////////////////////// PUBLIC SECTION //////////////////////////////////////////
     /////////////////////////////////////////////////////////////////////////////{{{
@@ -646,7 +653,8 @@ liberator.History = function () //{{{
             if (!history)
                 load();
 
-                return liberator.completion.filterURLArray(history, filter);
+            return liberator.completion.filterURLArray(cachedHistory, filter).concat(
+                   liberator.completion.filterURLArray(history, filter));
         },
 
         // the history is automatically added to the Places global history
@@ -656,11 +664,16 @@ liberator.History = function () //{{{
             if (!history)
                 load();
 
-            history = history.filter(function (elem) {
-                return elem[0] != url;
-            });
+            // don' let cachedHistory grow too large
+            if (cachedHistory.length > 1000)
+            {
+                history = cachedHistory.concat(history);
+                cachedHistory = [];
+            }
+            else
+                cachedHistory = cachedHistory.filter(function (elem) { return elem[0] != url; });
 
-            history.unshift([url, title]);
+            cachedHistory.unshift([url, title || "[No title]"]);
             return true;
         },
 
@@ -734,7 +747,6 @@ liberator.History = function () //{{{
             }
             else
             {
-
                 var list = ":" + liberator.util.escapeHTML(liberator.commandline.getCommand()) + "<br/>" +
                            "<table><tr align=\"left\" class=\"hl-Title\"><th>title</th><th>URL</th></tr>";
                 for (var i = 0; i < items.length; i++)
@@ -749,7 +761,6 @@ liberator.History = function () //{{{
                 liberator.commandline.echo(list, liberator.commandline.HL_NORMAL, liberator.commandline.FORCE_MULTILINE);
             }
         }
-
     };
     //}}}
 }; //}}}
@@ -773,6 +784,7 @@ liberator.QuickMarks = function () //{{{
     /////////////////////////////////////////////////////////////////////////////}}}
     ////////////////////// MAPPINGS ////////////////////////////////////////////////
     /////////////////////////////////////////////////////////////////////////////{{{
+
     var modes = liberator.config.browserModes || [liberator.modes.NORMAL];
 
     liberator.mappings.add(modes,
@@ -834,20 +846,15 @@ liberator.QuickMarks = function () //{{{
         "Mark a URL with a letter for quick access",
         function (args)
         {
-            if (!args)
-            {
-                liberator.echoerr("E471: Argument required");
-                return;
-            }
-
-            var matches = args.match(/^([a-zA-Z0-9])(?:\s+(.+))?$/);
+            var matches = args.string.match(/^([a-zA-Z0-9])(?:\s+(.+))?$/);
             if (!matches)
                 liberator.echoerr("E488: Trailing characters");
             else if (!matches[2])
                 liberator.quickmarks.add(matches[1], liberator.buffer.URL);
             else
                 liberator.quickmarks.add(matches[1], matches[2]);
-        });
+        },
+        { argCount: "+" });
 
     liberator.commands.add(["qmarks"],
         "Show all QuickMarks",
