@@ -105,7 +105,7 @@ liberator.Buffer = function () //{{{
     function findScrollableWindow()
     {
         var win = window.document.commandDispatcher.focusedWindow;
-        if (win.scrollMaxX > 0 || win.scrollMaxY > 0)
+        if (win && (win.scrollMaxX > 0 || win.scrollMaxY > 0))
             return win;
 
         win = window.content;
@@ -142,7 +142,12 @@ liberator.Buffer = function () //{{{
     ////////////////////// OPTIONS /////////////////////////////////////////////////
     /////////////////////////////////////////////////////////////////////////////{{{
 
-    liberator.options.add(["fullscreen", "fs"], "Show the current window fullscreen", "boolean", false,
+    // override this stupid pref, because otherwise zoom is lost after switching tabs
+    liberator.options.setPref("browser.zoom.siteSpecific", false);
+
+    liberator.options.add(["fullscreen", "fs"],
+        "Show the current window fullscreen",
+        "boolean", false,
         {
             setter: function (value) { window.fullScreen = value; },
             getter: function () { return window.fullScreen; }
@@ -173,15 +178,37 @@ liberator.Buffer = function () //{{{
         "Show the destination of the link under the cursor in the status bar",
         "number", 1,
         {
-            validator: function (value) { return (value >= 0 && value <= 2); }
+            validator: function (value) { return (value >= 0 && value <= 2); },
+            completer: function (filter)
+            {
+                return [
+                  ["0", "Don't show link destination"],
+                  ["1", "Show the link in the status line"],
+                  ["2", "Show the link in the command line"]
+                ];
+            }
         });
 
     liberator.options.add(["usermode", "um"],
         "Show current website with a minimal style sheet to make it easily accessible",
         "boolean", false,
         {
-            setter: function (value) { getMarkupDocumentViewer().authorStyleDisabled = value; },
-            getter: function () { return getMarkupDocumentViewer().authorStyleDisabled; },
+            setter: function (value)
+            {
+                try
+                {
+                    getMarkupDocumentViewer().authorStyleDisabled = value;
+                }
+                catch (e) {}
+            },
+            getter: function ()
+            {
+                try
+                {
+                    return getMarkupDocumentViewer().authorStyleDisabled;
+                }
+                catch (e) {}
+            }
         });
 
     /////////////////////////////////////////////////////////////////////////////}}}
@@ -419,15 +446,54 @@ liberator.Buffer = function () //{{{
 
     liberator.commands.add(["ha[rdcopy]"],
         "Print current document",
-        function () { getBrowser().contentWindow.print(); });
+        function (args, special)
+        {
+            // TODO: Hide the progress bar ("print.show_print_progress") when
+            // the special version is called?
+            var value = liberator.options.getPref("print.always_print_silent");
+            liberator.options.setPref("print.always_print_silent", special);
+
+            getBrowser().contentWindow.print();
+
+            liberator.options.setPref("print.always_print_silent", value);
+        });
 
     liberator.commands.add(["pa[geinfo]"],
         "Show various page information",
-        function () { liberator.buffer.showPageInfo(true); });
+        function ()
+        {
+            liberator.buffer.showPageInfo(true);
+        },
+        { argCount: "0" });
+
+    liberator.commands.add(["pagest[yle]"],
+        "Select the author style sheet to apply",
+        function (args)
+        {
+            var titles = liberator.buffer.alternateStyleSheets.map(function (stylesheet) { return stylesheet.title; });
+
+            if (args && !titles.some(function (title) { return title == args; }))
+            {
+                liberator.echoerr("Exxx: No matching stylesheet");
+                return;
+            }
+
+            if (liberator.options["usermode"])
+                liberator.options["usermode"] = false;
+
+            stylesheetSwitchAll(window.content, args);
+        },
+        {
+            completer: function (filter) { return liberator.completion.stylesheet(filter); }
+        });
 
     liberator.commands.add(["re[load]"],
         "Reload current page",
-        function (args, special) { liberator.tabs.reload(getBrowser().mCurrentTab, special); });
+        function (args, special)
+        {
+            liberator.tabs.reload(getBrowser().mCurrentTab, special);
+        },
+        { argCount: "0" });
 
     liberator.commands.add(["sav[eas]", "w[rite]"],
         "Save current document to disk",
@@ -447,7 +513,11 @@ liberator.Buffer = function () //{{{
 
     liberator.commands.add(["st[op]"],
         "Stop loading",
-        function () { BrowserStop(); });
+        function ()
+        {
+            BrowserStop();
+        },
+        { argCount: "0" });
 
     liberator.commands.add(["vie[wsource]"],
         "View source code of current document",
@@ -498,6 +568,18 @@ liberator.Buffer = function () //{{{
 
     return {
 
+        get alternateStyleSheets()
+        {
+            var stylesheets = getAllStyleSheets(window.content);
+
+            // TODO: how should we handle duplicate titles?
+            stylesheets = stylesheets.filter(function (stylesheet) {
+                return !(!/^(screen|all|)$/i.test(stylesheet.media.mediaText) || /^\s*$/.test(stylesheet.title))
+            });
+
+            return stylesheets;
+        },
+
         // 0 if loading, 1 if loaded or 2 if load failed
         get loaded()
         {
@@ -526,8 +608,6 @@ liberator.Buffer = function () //{{{
 
         get URL()
         {
-            // TODO: .URL is not defined for XUL documents
-            //return window.content.document.URL;
             return window.content.document.location.href;
         },
 
@@ -585,6 +665,29 @@ liberator.Buffer = function () //{{{
             return result;
         },
 
+        // in contrast to vim, returns the selection if one is made,
+        // otherwise tries to guess the current word unter the text cursor
+        // NOTE: might change the selection
+        getCurrentWord: function ()
+        {
+            var selection = window.content.getSelection().toString();
+
+            if (!selection)
+            {
+                var selectionController = getBrowser().docShell
+                    .QueryInterface(Components.interfaces.nsIInterfaceRequestor)
+                    .getInterface(Components.interfaces.nsISelectionDisplay)
+                    .QueryInterface(Components.interfaces.nsISelectionController);
+
+                selectionController.setCaretEnabled(true);
+                selectionController.wordMove(false, false);
+                selectionController.wordMove(true, true);
+                selection = window.content.getSelection().toString();
+            }
+
+            return selection;
+        },
+
         // quick function to get elements inside the document reliably
         // argument "args" is something like: @id='myid' or @type='text' (don't forget the quotes around myid)
         getElement: function (args, index)
@@ -592,50 +695,6 @@ liberator.Buffer = function () //{{{
             return liberator.buffer.evaluateXPath("//*[" + (args || "") + "]").snapshotItem(index || 0);
         },
 
-        // artificially "clicks" a link in order to open it
-        followLink: function (elem, where)
-        {
-            var doc = window.content.document;
-            var view = window.document.defaultView;
-            var offsetX = 1;
-            var offsetY = 1;
-
-            var localName = elem.localName.toLowerCase();
-            if (localName == "frame" || localName == "iframe") // broken?
-            {
-                elem.contentWindow.focus();
-                return false;
-            }
-            else if (localName == "area") // for imagemap
-            {
-                var coords = elem.getAttribute("coords").split(",");
-                offsetX = Number(coords[0]) + 1;
-                offsetY = Number(coords[1]) + 1;
-            }
-
-            var newTab = false, newWindow = false;
-            switch (where)
-            {
-                case liberator.NEW_TAB:
-                case liberator.NEW_BACKGROUND_TAB:
-                    newTab = true;
-                    break;
-                case liberator.NEW_WINDOW:
-                    newWindow = true;
-                    break;
-                default:
-                    liberator.log("Invalid where argument for followLink()");
-            }
-
-            elem.focus();
-
-            var evt = doc.createEvent("MouseEvents");
-            evt.initMouseEvent("mousedown", true, true, view, 1, offsetX, offsetY, 0, 0, /*ctrl*/ newTab, /*event.altKey*/0, /*event.shiftKey*/ newWindow, /*event.metaKey*/ newTab, 0, null);
-            elem.dispatchEvent(evt);
-            evt.initMouseEvent("click", true, true, view, 1, offsetX, offsetY, 0, 0, /*ctrl*/ newTab, /*event.altKey*/0, /*event.shiftKey*/ newWindow, /*event.metaKey*/ newTab, 0, null);
-            elem.dispatchEvent(evt);
-        },
-
         // more advanced than a simple elem.focus() as it also works for iframes
         // and image maps
         // TODO: merge with followLink()?
@@ -668,6 +727,149 @@ liberator.Buffer = function () //{{{
             elem.dispatchEvent(evt);
         },
 
+        followDocumentRelationship: function (relationship)
+        {
+            function followFrameRelationship(relationship, parsedFrame)
+            {
+                var regexps;
+                var relText;
+                var patternText;
+                var revString;
+                switch (relationship)
+                {
+                    case "next":
+                        regexps = liberator.options["nextpattern"].split(",");
+                        revString = "previous";
+                        break;
+                    case "previous":
+                        // TODO: accept prev\%[ious]
+                        regexps = liberator.options["previouspattern"].split(",");
+                        revString = "next";
+                        break;
+                    default:
+                        liberator.echoerr("Bad document relationship: " + relationship);
+                }
+
+                relText = new RegExp(relationship, "i");
+                revText = new RegExp(revString, "i");
+                var elems = parsedFrame.document.getElementsByTagName("link");
+                // links have higher priority than normal <a> hrefs
+                for (var i = 0; i < elems.length; i++)
+                {
+                    if (relText.test(elems[i].rel) || revText.test(elems[i].rev))
+                    {
+                            liberator.open(elems[i].href);
+                            return true;
+                    }
+                }
+
+                // no links? ok, look for hrefs
+                elems = parsedFrame.document.getElementsByTagName("a");
+                for (var i = 0; i < elems.length; i++)
+                {
+                    if (relText.test(elems[i].rel) || revText.test(elems[i].rev))
+                    {
+                        liberator.buffer.followLink(elems[i], liberator.CURRENT_TAB);
+                        return true;
+                    }
+                }
+
+                for (var pattern = 0; pattern < regexps.length; pattern++)
+                {
+                    patternText = new RegExp(regexps[pattern], "i");
+                    for (var i = 0; i < elems.length; i++)
+                    {
+                        if (patternText.test(elems[i].textContent))
+                        {
+                            liberator.buffer.followLink(elems[i], liberator.CURRENT_TAB);
+                            return true;
+                        }
+                        else
+                        {
+                            // images with alt text being href
+                            var children = elems[i].childNodes;
+                            for (var j = 0; j < children.length; j++)
+                            {
+                                if (patternText.test(children[j].alt))
+                                {
+                                    liberator.buffer.followLink(elems[i], liberator.CURRENT_TAB);
+                                    return true;
+                                }
+                            }
+                        }
+                    }
+                }
+                return false;
+            }
+
+            var retVal;
+            if (window.content.frames.length != 0)
+            {
+                retVal = followFrameRelationship(relationship, window.content);
+                if (!retVal)
+                {
+                    // only loop through frames if the main content didnt match
+                    for (var i = 0; i < window.content.frames.length; i++)
+                    {
+                        retVal = followFrameRelationship(relationship, window.content.frames[i]);
+                        if (retVal)
+                            break;
+                    }
+                }
+            }
+            else
+            {
+                retVal = followFrameRelationship(relationship, window.content);
+            }
+
+            if (!retVal)
+                liberator.beep();
+        },
+
+        // artificially "clicks" a link in order to open it
+        followLink: function (elem, where)
+        {
+            var doc = window.content.document;
+            var view = window.document.defaultView;
+            var offsetX = 1;
+            var offsetY = 1;
+
+            var localName = elem.localName.toLowerCase();
+            if (localName == "frame" || localName == "iframe") // broken?
+            {
+                elem.contentWindow.focus();
+                return false;
+            }
+            else if (localName == "area") // for imagemap
+            {
+                var coords = elem.getAttribute("coords").split(",");
+                offsetX = Number(coords[0]) + 1;
+                offsetY = Number(coords[1]) + 1;
+            }
+
+            var newTab = false, newWindow = false;
+            switch (where)
+            {
+                case liberator.NEW_TAB:
+                case liberator.NEW_BACKGROUND_TAB:
+                    newTab = true;
+                    break;
+                case liberator.NEW_WINDOW:
+                    newWindow = true;
+                    break;
+                default:
+                    liberator.log("Invalid where argument for followLink()", 0);
+            }
+
+            elem.focus();
+
+            var evt = doc.createEvent("MouseEvents");
+            evt.initMouseEvent("mousedown", true, true, view, 1, offsetX, offsetY, 0, 0, /*ctrl*/ newTab, /*event.altKey*/0, /*event.shiftKey*/ newWindow, /*event.metaKey*/ newTab, 0, null);
+            elem.dispatchEvent(evt);
+            evt.initMouseEvent("click", true, true, view, 1, offsetX, offsetY, 0, 0, /*ctrl*/ newTab, /*event.altKey*/0, /*event.shiftKey*/ newWindow, /*event.metaKey*/ newTab, 0, null);
+            elem.dispatchEvent(evt);
+        },
+
         saveLink: function (elem, skipPrompt)
         {
             var doc  = elem.ownerDocument;
@@ -687,29 +889,6 @@ liberator.Buffer = function () //{{{
             }
         },
 
-        // in contrast to vim, returns the selection if one is made,
-        // otherwise tries to guess the current word unter the text cursor
-        // NOTE: might change the selection
-        getCurrentWord: function ()
-        {
-            var selection = window.content.getSelection().toString();
-
-            if (!selection)
-            {
-                var selectionController = getBrowser().docShell
-                    .QueryInterface(Components.interfaces.nsIInterfaceRequestor)
-                    .getInterface(Components.interfaces.nsISelectionDisplay)
-                    .QueryInterface(Components.interfaces.nsISelectionController);
-
-                selectionController.setCaretEnabled(true);
-                selectionController.wordMove(false, false);
-                selectionController.wordMove(true, true);
-                selection = window.content.getSelection().toString();
-            }
-
-            return selection;
-        },
-
         scrollBottom: function ()
         {
             scrollToPercentiles(-1, 100);
@@ -873,28 +1052,6 @@ liberator.Buffer = function () //{{{
             setTimeout(function () { doc.body.removeChild(indicator); }, 500);
         },
 
-        // XXX: probably remove this method/functionality
-        // updates the buffer preview in place only if list is visible
-        updateBufferList: function ()
-        {
-            if (!liberator.bufferwindow.visible())
-                return;
-
-            var items = liberator.completion.buffer("")[1];
-            liberator.bufferwindow.show(items);
-            liberator.bufferwindow.selectItem(getBrowser().mTabContainer.selectedIndex);
-        },
-
-        zoomIn: function (steps, fullZoom)
-        {
-            bumpZoomLevel(steps, fullZoom);
-        },
-
-        zoomOut: function (steps, fullZoom)
-        {
-            bumpZoomLevel(-steps, fullZoom);
-        },
-
         // similar to pageInfo
         // TODO: print more useful information, just like the DOM inspector
         showElementInfo: function (elem)
@@ -993,7 +1150,7 @@ liberator.Buffer = function () //{{{
                 {
                     cacheEntryDescriptor = ftpCacheSession.openCacheEntry(cacheKey, ACCESS_READ, false);
                 }
-                catch (e) { }
+                catch (e) {}
             }
 
             var pageSize = []; // [0] bytes; [1] kbytes
@@ -1035,7 +1192,7 @@ liberator.Buffer = function () //{{{
 
             var lastModVerbose = new Date(window.content.document.lastModified).toLocaleString();
             var lastMod = new Date(window.content.document.lastModified).toLocaleFormat("%x %X");
-            // FIXME: probably unportable across differnet language versions
+            // FIXME: probably unportable across different language versions
             if (lastModVerbose == "Invalid Date" || new Date(window.content.document.lastModified).getFullYear() == 1970)
                 lastModVerbose = lastMod = null;
 
@@ -1152,105 +1309,6 @@ liberator.Buffer = function () //{{{
             liberator.echo(pageInfoText, liberator.commandline.FORCE_MULTILINE);
         },
 
-        followDocumentRelationship: function (relationship)
-        {
-            function followFrameRelationship(relationship, parsedFrame)
-            {
-                var regexps;
-                var relText;
-                var patternText;
-                var revString;
-                switch (relationship)
-                {
-                    case "next":
-                        regexps = liberator.options["nextpattern"].split(",");
-                        revString = "previous";
-                        break;
-                    case "previous":
-                        // TODO: accept prev\%[ious]
-                        regexps = liberator.options["previouspattern"].split(",");
-                        revString = "next";
-                        break;
-                    default:
-                        liberator.echoerr("Bad document relationship: " + relationship);
-                }
-
-                relText = new RegExp(relationship, "i");
-                revText = new RegExp(revString, "i");
-                var elems = parsedFrame.document.getElementsByTagName("link");
-                // links have higher priority than normal <a> hrefs
-                for (var i = 0; i < elems.length; i++)
-                {
-                    if (relText.test(elems[i].rel) || revText.test(elems[i].rev))
-                    {
-                            liberator.open(elems[i].href);
-                            return true;
-                    }
-                }
-
-                // no links? ok, look for hrefs
-                elems = parsedFrame.document.getElementsByTagName("a");
-                for (var i = 0; i < elems.length; i++)
-                {
-                    if (relText.test(elems[i].rel) || revText.test(elems[i].rev))
-                    {
-                        liberator.buffer.followLink(elems[i], liberator.CURRENT_TAB);
-                        return true;
-                    }
-                }
-
-                for (var pattern = 0; pattern < regexps.length; pattern++)
-                {
-                    patternText = new RegExp(regexps[pattern], "i");
-                    for (var i = 0; i < elems.length; i++)
-                    {
-                        if (patternText.test(elems[i].textContent))
-                        {
-                            liberator.buffer.followLink(elems[i], liberator.CURRENT_TAB);
-                            return true;
-                        }
-                        else
-                        {
-                            // images with alt text being href
-                            var children = elems[i].childNodes;
-                            for (var j = 0; j < children.length; j++)
-                            {
-                                if (patternText.test(children[j].alt))
-                                {
-                                    liberator.buffer.followLink(elems[i], liberator.CURRENT_TAB);
-                                    return true;
-                                }
-                            }
-                        }
-                    }
-                }
-                return false;
-            }
-
-            var retVal;
-            if (window.content.frames.length != 0)
-            {
-                retVal = followFrameRelationship(relationship, window.content);
-                if (!retVal)
-                {
-                    // only loop through frames if the main content didnt match
-                    for (var i = 0; i < window.content.frames.length; i++)
-                    {
-                        retVal = followFrameRelationship(relationship, window.content.frames[i]);
-                        if (retVal)
-                            break;
-                    }
-                }
-            }
-            else
-            {
-                retVal = followFrameRelationship(relationship, window.content);
-            }
-
-            if (!retVal)
-                liberator.beep();
-        },
-
         viewSelectionSource: function ()
         {
             // copied (and tuned somebit) from browser.jar -> nsContextMenu.js
@@ -1281,21 +1339,31 @@ liberator.Buffer = function () //{{{
                 // TODO: save return value in v:shell_error
                 var newThread = Components.classes["@mozilla.org/thread-manager;1"].getService().newThread(0);
                 var editor = liberator.options["editor"];
-                var args = editor.split(" "); // FIXME: too simple
+                var args = liberator.commands.parseArgs(editor, [], "*", true).arguments;
                 if (args.length < 1)
                 {
-                    liberator.echoerr("no editor specified");
+                    liberator.echoerr("No editor specified");
                     return;
                 }
 
                 var prog = args.shift();
-                args.push(url)
+                args.push(url);
                 liberator.callFunctionInThread(newThread, liberator.io.run, [prog, args, true]);
             }
             else
             {
-                liberator.open("view-source:" + url)
+                liberator.open("view-source:" + url);
             }
+        },
+
+        zoomIn: function (steps, fullZoom)
+        {
+            bumpZoomLevel(steps, fullZoom);
+        },
+
+        zoomOut: function (steps, fullZoom)
+        {
+            bumpZoomLevel(-steps, fullZoom);
         }
     };
     //}}}
@@ -1479,24 +1547,21 @@ liberator.Marks = function () //{{{
         "Mark current location within the web page",
         function (args)
         {
-            if (!args)
-            {
-                liberator.echoerr("E471: Argument required");
-                return;
-            }
-            if (args.length > 1)
+            var mark = args.arguments[0];
+            if (mark.length > 1)
             {
                 liberator.echoerr("E488: Trailing characters");
                 return;
             }
-            if (!/[a-zA-Z]/.test(args))
+            if (!/[a-zA-Z]/.test(mark))
             {
                 liberator.echoerr("E191: Argument must be a letter or forward/backward quote");
                 return;
             }
 
-            liberator.marks.add(args);
-        });
+            liberator.marks.add(mark);
+        },
+        { argCount: "1" });
 
     liberator.commands.add(["marks"],
         "Show all location marks of current web page",
@@ -1526,7 +1591,7 @@ liberator.Marks = function () //{{{
 
             if (win.document.body.localName.toLowerCase() == "frameset")
             {
-                liberator.echoerr("marks support for frameset pages not implemented yet");
+                liberator.echoerr("Marks support for frameset pages not implemented yet");
                 return;
             }
 
