@@ -34,74 +34,96 @@ liberator.Completion = function () //{{{
 
     // the completion substrings, used for showing the longest common match
     var substrings = [];
+    var urlResultsCache = null;
+    var urlCompletionCache = [];
 
     // function uses smartcase
     // list = [ [['com1', 'com2'], 'text'], [['com3', 'com4'], 'text'] ]
     function buildLongestCommonSubstring(list, filter)
     {
         var filtered = [];
+
         var ignorecase = false;
         if (filter == filter.toLowerCase())
             ignorecase = true;
 
+        var longest = false;
+        if (liberator.options["wildmode"].indexOf("longest") >= 0)
+            longest = true;
+
         for (var i = 0; i < list.length; i++)
         {
-            for (var j = 0; j < list[i][0].length; j++)
+            var complist = list[i][0] instanceof Array ? list[i][0] : [list[i][0]];
+            for (var j = 0; j < complist.length; j++)
             {
-                var item = list[i][0][j];
+                var item = complist[j];
                 if (ignorecase)
                     item = item.toLowerCase();
 
                 if (item.indexOf(filter) == -1)
                     continue;
 
-                if (substrings.length == 0)
+                filtered.push([complist[j], list[i][1]]);
+
+                if (longest)
                 {
-                    var lastIndex = item.lastIndexOf(filter);
-                    var length = item.length;
-                    for (var k = item.indexOf(filter); k != -1 && k <= lastIndex; k = item.indexOf(filter, k + 1))
+                    if (substrings.length == 0)
                     {
-                        for (var l = k + filter.length; l <= length; l++)
-                            substrings.push(list[i][0][j].substring(k, l));
+                        var lastIndex = item.lastIndexOf(filter);
+                        var length = item.length;
+                        for (var k = item.indexOf(filter); k != -1 && k <= lastIndex; k = item.indexOf(filter, k + 1))
+                        {
+                            for (var l = k + filter.length; l <= length; l++)
+                                substrings.push(complist[j].substring(k, l));
+                        }
+                    }
+                    else
+                    {
+                        substrings = substrings.filter(function ($_) {
+                            return complist[j].indexOf($_) >= 0;
+                        });
                     }
                 }
-                else
-                {
-                    substrings = substrings.filter(function ($_) {
-                        return list[i][0][j].indexOf($_) >= 0;
-                    });
-                }
-                filtered.push([list[i][0][j], list[i][1]]);
                 break;
             }
         }
         return filtered;
     }
 
-    // this function is case sensitive and should be documented about input and output ;)
+    // this function is case sensitive
     function buildLongestStartingSubstring(list, filter)
     {
         var filtered = [];
+
+        var longest = false;
+        if (liberator.options["wildmode"].indexOf("longest") >= 0)
+            longest = true;
+
         for (var i = 0; i < list.length; i++)
         {
-            for (var j = 0; j < list[i][0].length; j++)
+            var complist = list[i][0] instanceof Array ? list[i][0] : [list[i][0]];
+            for (var j = 0; j < complist.length; j++)
             {
-                if (list[i][0][j].indexOf(filter) != 0)
+                if (complist[j].indexOf(filter) != 0)
                     continue;
 
-                if (substrings.length == 0)
+                filtered.push([complist[j], list[i][1]]);
+
+                if (longest)
                 {
-                    var length = list[i][0][j].length;
-                    for (var k = filter.length; k <= length; k++)
-                        substrings.push(list[i][0][j].substring(0, k));
+                    if (substrings.length == 0)
+                    {
+                        var length = complist[j].length;
+                        for (var k = filter.length; k <= length; k++)
+                            substrings.push(complist[j].substring(0, k));
+                    }
+                    else
+                    {
+                        substrings = substrings.filter(function ($_) {
+                            return complist[j].indexOf($_) == 0;
+                        });
+                    }
                 }
-                else
-                {
-                    substrings = substrings.filter(function ($_) {
-                        return list[i][0][j].indexOf($_) == 0;
-                    });
-                }
-                filtered.push([list[i][0][j], list[i][1]]);
                 break;
             }
         }
@@ -129,59 +151,25 @@ liberator.Completion = function () //{{{
             }
             return longest;
         },
-        autocommands: function (filter)
+
+        // generic filter function, also builds substrings needed
+        // for :set wildmode=list:longest, if necessary
+        filter: function (array, filter, matchFromBeginning)
         {
-            substrings = [];
-            var nodes = [
-                ["BrowserExit",    "when firefox exits"],
-                ["BrowserRestart", "when firefox restarts"],
-                ["PageLoad",       "when a page gets (re)loaded/opened"]
-            ];
-
             if (!filter)
-                return [0, nodes];
+                return array;
 
-            var mapped = nodes.map(function (node) {
-                return [[node[0]], node[1]];
-            });
-
-            return [0, buildLongestCommonSubstring(mapped, filter)];
-        },
-        dialog: function (filter)
-        {
-            substrings = [];
-            var nodes = liberator.config.dialogs || [];
-
-            if (!filter)
-                return [0, nodes];
-
-            var mapped = nodes.map(function (node) {
-                return [[node[0]], node[1]];
-            });
-
-            return [0, buildLongestCommonSubstring(mapped, filter)];
-        },
-
-        macros: function (filter)
-        {
-            var macros = [];
-            var tmp = liberator.events.getMacros();
-            for (var item in tmp)
-                macros.push([item, tmp[item]]);
-
-            if (!filter)
-                return [0, macros];
-
-            var mapped = macros.map(function (macro) {
-                return [[macro[0]], macro[1]];
-            });
-            return [0, buildLongestCommonSubstring(mapped, filter)];
+            if (matchFromBeginning)
+                return buildLongestStartingSubstring(array, filter);
+            else
+                return buildLongestCommonSubstring(array, filter);
         },
 
+        // XXX: Move to bookmarks.js?
         searchEngineSuggest: function (filter, engineAliases)
         {
             if (!filter)
-                return [0,null];
+                return [0, []];
 
         	var engineList = (engineAliases || liberator.options["suggestengines"]).split(",");
         	var responseType = "application/x-suggestions+json";
@@ -202,24 +190,25 @@ liberator.Completion = function () //{{{
             	if (engine && engine.supportsResponseType(responseType))
                     queryURI = engine.getSubmission(query, responseType).uri.asciiSpec;
                 else
-                    return;
+                    return [0, []];
 
             	var xhr = new XMLHttpRequest();
             	xhr.open("GET", queryURI, false);
             	xhr.send(null);
 
-                var json = Components.classes["@mozilla.org/dom/json;1"].createInstance(Components.interfaces.nsIJSON);
+                var json = Components.classes["@mozilla.org/dom/json;1"]
+                           .createInstance(Components.interfaces.nsIJSON);
                 var results = json.decode(xhr.responseText)[1];
                 if (!results)
-                    return;
+                    return [0, []];
 
             	results.forEach(function (item)
             	{
                     // make sure we receive strings, otherwise a man-in-the-middle attack
                     // could return objects which toString() method could be called to
                     // execute untrusted code
-                    if(typeof(item) != "string")
-                        return;
+                    if (typeof item != "string")
+                        return [0, []];
 
             	    completions.push([(matches ? matches[1] : "") + item, engine.name + " suggestion"]);
             	});
@@ -236,10 +225,8 @@ liberator.Completion = function () //{{{
         url: function (filter, complete)
         {
             var completions = [];
-            substrings = [];
-
             var start = 0;
-            var skip = filter.match(/^(.*,\s+)(.*)/); // start after the last ", "
+            var skip = filter.match("^(.*" + liberator.options["urlseparator"] + ")(.*)"); // start after the last 'urlseparator'
             if (skip)
             {
                 start += skip[1].length;
@@ -247,6 +234,7 @@ liberator.Completion = function () //{{{
             }
 
             var cpt = complete || liberator.options["complete"];
+            var autoCompletions = liberator.options["wildoptions"].indexOf("auto") >= 0;
             var suggestEngineAlias = liberator.options["suggestengines"] || "google";
             // join all completion arrays together
             for (var i = 0; i < cpt.length; i++)
@@ -255,40 +243,63 @@ liberator.Completion = function () //{{{
                     completions = completions.concat(this.search(filter)[1]);
                 else if (cpt[i] == "f")
                     completions = completions.concat(this.file(filter, false)[1]);
-                else if (cpt[i] == "b")
+                else if (!autoCompletions && cpt[i] == "b")
                     completions = completions.concat(liberator.bookmarks.get(filter));
-                else if (cpt[i] == "h")
+                else if (!autoCompletions && cpt[i] == "h")
                     completions = completions.concat(liberator.history.get(filter));
                 else if (cpt[i] == "S")
                     completions = completions.concat(this.searchEngineSuggest(filter, suggestEngineAlias)[1]);
+                else if (autoCompletions && cpt[i] == "l") // add completions like Firefox's smart location bar
+                {
+                    var completionService = Components.classes["@mozilla.org/browser/global-history;2"].
+                                            getService(Components.interfaces.nsIAutoCompleteSearch);
+                    completionService.startSearch(filter, "", urlResultsCache, {
+                        onSearchResult: function (search, result) {
+                            //if (result.searchResult != result.RESULT_SUCCESS)
+                            //    return;
+                            //liberator.log(result.searchResult);
+                            //var res = "";// + util.objectToString(result) + "\n---\n";
+                            //liberator.log(result.matchCount + " matches: " + result.searchResult);
+                            var comp = [];
+                            //if (result.searchResult == result.RESULT_SUCCESS)
+                            //    urlResultsCache = result;
+
+                            for (var i = 0; i < result.matchCount; i++)
+                            {
+                                comp.push([result.getValueAt(i), result.getCommentAt(i)]);
+                            }
+                            urlCompletionCache = comp;
+                            if (comp.length > 0 || result.searchResult == result.RESULT_SUCCESS)
+                                liberator.commandline.setCompletions(completions.concat(comp));
+                        }
+                    });
+                }
             }
 
-            return [start, completions];
+            return [start, completions.concat(urlCompletionCache)];
         },
 
         search: function (filter)
         {
             var engines = liberator.bookmarks.getSearchEngines().concat(liberator.bookmarks.getKeywords());
+            return [0, this.filter(engines, filter)];
+        },
 
-            if (!filter)
-                return [0, engines];
+        stylesheet: function (filter)
+        {
+            var stylesheets = getAllStyleSheets(window.content);
 
-            var mapped = engines.map(function (engine) {
-                return [[engine[0]], engine[1]];
+            stylesheets = liberator.buffer.alternateStyleSheets.map(function (stylesheet) {
+                return [stylesheet.title, stylesheet.href || "inline"];
             });
-            return [0, buildLongestCommonSubstring(mapped, filter)];
+
+            return [0, this.filter(stylesheets, filter)];
         },
 
         // TODO: support file:// and \ or / path separators on both platforms
-        // TODO: sort directories first
         // if "tail" is true, only return names without any directory components
         file: function (filter, tail)
         {
-            // this is now also used as part of the url completion, so the
-            // substrings shouldn't be cleared for that case
-            if (!arguments[1])
-                substrings = [];
-
             var dir = "", compl = "";
             var matches = filter.match(/^(.*[\/\\])?(.*?)$/);
             if (matches)
@@ -296,13 +307,17 @@ liberator.Completion = function () //{{{
                 dir = matches[1] || ""; // "" is expanded inside readDirectory to the current dir
                 compl = matches[2] || "";
             }
-            var files = [], mapped = [];
 
+            var files = [], mapped = [];
             try
             {
                 files = liberator.io.readDirectory(dir);
                 mapped = files.map(function (file) {
-                    return [[tail ? file.leafName : (dir + file.leafName)], file.isDirectory() ? "Directory" : "File"];
+                    return [tail ? file.leafName : (dir + file.leafName), file.isDirectory() ? "Directory" : "File"];
+                }).sort(function (a, b) {
+                    return a[0] < b[0] ? -1 : a[0] > b[0] ? 1 : 0
+                }).sort(function (a, b) {
+                    return a[1] < b[1] ? -1 : a[1] > b[1] ? 1 : 0
                 });
             }
             catch (e)
@@ -316,45 +331,8 @@ liberator.Completion = function () //{{{
                 return [0, buildLongestStartingSubstring(mapped, filter)];
         },
 
-        help: function (filter)
-        {
-            var res = [];
-            // they are sorted by relevance, not alphabetically
-            var files = ["intro.html", "starting.html", "browsing.html", "buffer.html",
-                         "options.html", "tabs.html", "marks.html", "repeat.html",
-                         "autocommands.html", "developer.html", "various.html"];
-
-            for (var file in files)
-            {
-                try
-                {
-                    var xmlhttp = new XMLHttpRequest();
-                    xmlhttp.open("GET", "chrome://" + liberator.config.name.toLowerCase() + "/locale/" + files[file], false);
-                    xmlhttp.send(null);
-                }
-                catch (e)
-                {
-                    liberator.log("Error opening chrome://" + liberator.config.name.toLowerCase() + "/locale/" + files[file], 1);
-                    continue;
-                }
-                var doc = xmlhttp.responseXML;
-                var elems = doc.getElementsByClassName("tag");
-                for (var i = 0; i < elems.length; i++)
-                    res.push([elems[i].textContent, files[file]]);
-            }
-
-            if (!filter)
-                return [0, res];
-
-            var mapped = res.map(function (node) {
-                return [[node[0]], node[1]];
-            });
-            return [0, buildLongestCommonSubstring(mapped, filter)];
-        },
-
         command: function (filter)
         {
-            substrings = [];
             var completions = [];
 
             if (!filter)
@@ -370,142 +348,9 @@ liberator.Completion = function () //{{{
             return [0, buildLongestStartingSubstring(completions, filter)];
         },
 
-        mail: function (filter)
-        {
-            var completions = [];
-            var folders = liberator.mail.getFolders();
-            for (var folder in folders)
-            {
-                completions.push([folders[folder].server.prettyName + ": "
-                                  + folders[folder].name,
-                                 "Unread: " + folders[folder].getNumUnread(false)]);
-            }
-            if (!filter)
-                return [0, completions];
-            var mapped = completions.map(function (node) {
-                                         return [[node[0]], node[1]];
-                                         });
-            return [0, buildLongestCommonSubstring(mapped, filter)];
-        },
-
-        option: function (filter, special, unfiltered)
-        {
-            substrings = [];
-            var optionCompletions = [];
-            var prefix = filter.match(/^no|inv/) || "";
-
-            if (prefix)
-                filter = filter.replace(prefix, "");
-
-            // needed for help-completions, don't return [start, options], just options
-            // FIXME: doesn't belong here to be honest (rather v.options.get(filter)) --mst
-            if (unfiltered)
-            {
-                var options = [];
-                for (var option in liberator.options)
-                {
-                    if (prefix && option.type != "boolean")
-                        continue;
-                    options.push([option.names, option.description]);
-                }
-                return options;
-            }
-
-            if (special)
-            {
-                var prefs = Components.classes["@mozilla.org/preferences-service;1"]
-                    .getService(Components.interfaces.nsIPrefBranch);
-                var prefArray = prefs.getChildList("", { value: 0 });
-                prefArray.sort();
-
-                if (filter.length > 0 && filter.lastIndexOf("=") == filter.length - 1)
-                {
-                    for (var i = 0; i < prefArray.length; i++)
-                    {
-                        var name = prefArray[i];
-                        if (name.match("^" + filter.substr(0, filter.length - 1) + "$" ))
-                        {
-                            var value = liberator.options.getPref(name) + "";
-                            return [filter.length + 1, [[value, ""]]];
-                        }
-                    }
-                    return [0, []];
-                }
-
-                for (var i = 0; i < prefArray.length; i++)
-                {
-                    if (!filter)
-                        optionCompletions.push([prefArray[i], liberator.options.getPref(prefArray[i])]);
-                    else
-                        optionCompletions.push([[prefArray[i]], liberator.options.getPref(prefArray[i])]);
-                }
-
-
-                if (!filter)
-                    return [0, optionCompletions];
-
-                return [0, buildLongestCommonSubstring(optionCompletions, filter)];
-            }
-
-            if (!filter)
-            {
-                var options = [];
-                for (var option in liberator.options)
-                {
-                    if (prefix && option.type != "boolean")
-                        continue;
-                    options.push([prefix + option.name, option.description]);
-                }
-                return [0, options];
-            }
-            // check if filter ends with =, then complete current value
-            else if (filter.length > 0 && filter.lastIndexOf("=") == filter.length - 1)
-            {
-                filter = filter.substr(0, filter.length - 1);
-                for (var option in liberator.options)
-                {
-                    if (option.hasName(filter))
-                        return [filter.length + 1, [[option.value + "", ""]]];
-                }
-                return [0, optionCompletions];
-            }
-
-            // can't use b_l_s_s, since this has special requirements (the prefix)
-            var filterLength = filter.length;
-            for (var option in liberator.options)
-            {
-                if (prefix && option.type != "boolean")
-                    continue;
-
-                for (var j = 0; j < option.names.length; j++)
-                {
-                    if (option.names[j].indexOf(filter) != 0)
-                        continue;
-
-                    if (substrings.length == 0)
-                    {
-                        var length = option.names[j].length;
-                        for (var k = filterLength; k <= length; k++)
-                            substrings.push(prefix + option.names[j].substring(0, k));
-                    }
-                    else
-                    {
-                        substrings = substrings.filter(function ($_) {
-                            return option.names[j].indexOf($_) == 0;
-                        });
-                    }
-                    optionCompletions.push([prefix + option.names[j], option.description]);
-                    break;
-                }
-            }
-
-            return [0, optionCompletions];
-        },
-
         // FIXME: items shouldn't be [[[a], b]], but [[a, b]] and only mapped if at all for bLCS --mst
         buffer: function (filter)
         {
-            substrings = [];
             var items = [];
             var num = getBrowser().browsers.length;
             var title, url;
@@ -540,28 +385,8 @@ liberator.Completion = function () //{{{
             return [0, buildLongestCommonSubstring(items, filter)];
         },
 
-        sidebar: function (filter)
-        {
-            substrings = [];
-            var menu = document.getElementById("viewSidebarMenu");
-            var nodes = [];
-
-            for (var i = 0; i < menu.childNodes.length; i++)
-                nodes.push([menu.childNodes[i].label, ""]);
-
-            if (!filter)
-                return [0, nodes];
-
-            var mapped = nodes.map(function (node) {
-                return [[node[0]], node[1]];
-            });
-
-            return [0, buildLongestCommonSubstring(mapped, filter)];
-        },
-
         javascript: function (str)
         {
-            substrings = [];
             var matches = str.match(/^(.*?)(\s*\.\s*)?(\w*)$/);
             var objects = [];
             var filter = matches[3] || "";
@@ -755,10 +580,10 @@ liberator.Completion = function () //{{{
             return false;
         },
 
-        // FIXME: rename
-        // TODO: get completions for "nested" command lines like ":time :js <tab>" or ":tab :he<tab>"
-        exTabCompletion: function (str)
+        // provides completions for ex commands, including their arguments
+        ex: function (str)
         {
+            substrings = [];
             var [count, cmd, special, args] = liberator.commands.parseCommand(str);
             var completions = [];
             var start = 0;
