@@ -149,12 +149,12 @@ liberator.Editor = function () //{{{
     }
 
     // mode = "i" -> add :iabbrev, :iabclear and :iunabbrev commands
-    function addAbbreviationCommands(char, modeDescription)
+    function addAbbreviationCommands(ch, modeDescription)
     {
         var modeDescription = modeDescription ? " in " + modeDescription + " mode" : "";
-        var mode = char || "!";
+        var mode = ch || "!";
 
-        liberator.commands.add([char ? char + "a[bbrev]" : "ab[breviate]"],
+        liberator.commands.add([ch ? ch + "a[bbrev]" : "ab[breviate]"],
             "Abbreviate a key sequence" + modeDescription,
             function (args)
             {
@@ -172,13 +172,17 @@ liberator.Editor = function () //{{{
                     liberator.editor.listAbbreviations(mode, lhs);
             });
 
-        liberator.commands.add([char ? char + "una[bbrev]" : "una[bbreviate]"],
+        liberator.commands.add([ch ? ch + "una[bbrev]" : "una[bbreviate]"],
             "Remove an abbreviation" + modeDescription,
             function (args) { liberator.editor.removeAbbreviation(mode, args); });
 
-        liberator.commands.add([char + "abc[lear]"],
+        liberator.commands.add([ch + "abc[lear]"],
             "Remove all abbreviations" + modeDescription,
-            function (args) { liberator.editor.removeAllAbbreviations(mode); });
+            function ()
+            {
+                liberator.editor.removeAllAbbreviations(mode);
+            },
+            { argCount: "0" });
     }
 
     /////////////////////////////////////////////////////////////////////////////}}}
@@ -226,6 +230,11 @@ liberator.Editor = function () //{{{
     addMotionMap("y"); // yank
 
     // insert mode mappings
+    liberator.mappings.add(modes,
+        ["<C-o>", "<C-f>", "<C-g>", "<C-n>"],
+        "Ignore certain " + liberator.config.hostApplication + " key bindings",
+        function () { /*liberator.beep();*/ });
+
     liberator.mappings.add(modes,
         ["<C-w>"], "Delete previous word",
         function () { liberator.editor.executeCommand("cmd_deleteWordBackward", 1); });
@@ -261,6 +270,14 @@ liberator.Editor = function () //{{{
         ["<C-d>"], "Delete character to the right",
         function () { liberator.editor.executeCommand("cmd_deleteCharForward", 1); });
 
+    /*liberator.mappings.add(modes,
+        ["<C-Home>"], "Move cursor to beginning of text field",
+        function () { liberator.editor.executeCommand("cmd_moveTop", 1); });
+
+    liberator.mappings.add(modes,
+        ["<C-End>"], "Move cursor to end of text field",
+        function () { liberator.editor.executeCommand("cmd_moveBottom", 1); });*/
+
     liberator.mappings.add(modes,
         ["<S-Insert>"], "Insert clipboard/selection",
         function () { liberator.editor.pasteClipboard(); });
@@ -735,10 +752,10 @@ liberator.Editor = function () //{{{
                 return false;
 
             var editor = liberator.options["editor"];
-            var args = editor.split(" ");
+            var args = liberator.commands.parseArgs(editor, [], "*", true).arguments;
             if (args.length < 1)
             {
-                liberator.echoerr("no editor specified");
+                liberator.echoerr("No editor specified");
                 return false;
             }
 
@@ -762,7 +779,7 @@ liberator.Editor = function () //{{{
             }
 
             var prog = args.shift();
-            args.push(tmpfile.path)
+            args.push(tmpfile.path);
 
             if (textBox)
             {
