@@ -42,6 +42,14 @@ liberator.AutoCommands = function () //{{{
         throw StopIteration;
     }
 
+    /////////////////////////////////////////////////////////////////////////////}}}
+    ////////////////////// OPTIONS /////////////////////////////////////////////////
+    /////////////////////////////////////////////////////////////////////////////{{{
+
+    liberator.options.add(["focuscontent", "fc"],
+        "Try to stay in normal mode after loading a web page",
+        "boolean", false);
+
     /////////////////////////////////////////////////////////////////////////////}}}
     ////////////////////// COMMANDS ////////////////////////////////////////////////
     /////////////////////////////////////////////////////////////////////////////{{{
@@ -89,7 +97,10 @@ liberator.AutoCommands = function () //{{{
             }
         },
         {
-            completer: function (filter) { return liberator.completion.autocommands(filter); }
+            completer: function (filter)
+            {
+                return [0, liberator.completion.filter(liberator.config.autocommands || [], filter)];
+            }
         });
 
     /////////////////////////////////////////////////////////////////////////////}}}
@@ -218,7 +229,7 @@ liberator.AutoCommands = function () //{{{
         }
     };
     //}}}
-} //}}}
+}; //}}}
 
 liberator.Events = function () //{{{
 {
@@ -241,17 +252,14 @@ liberator.Events = function () //{{{
             tabcontainer.addEventListener("TabMove", function (event)
             {
                 liberator.statusline.updateTabCount();
-                liberator.buffer.updateBufferList();
             }, false);
             tabcontainer.addEventListener("TabOpen", function (event)
             {
                 liberator.statusline.updateTabCount();
-                liberator.buffer.updateBufferList();
             }, false);
             tabcontainer.addEventListener("TabClose", function (event)
             {
                 liberator.statusline.updateTabCount();
-                liberator.buffer.updateBufferList();
             }, false);
             tabcontainer.addEventListener("TabSelect", function (event)
             {
@@ -260,10 +268,10 @@ liberator.Events = function () //{{{
                 liberator.commandline.clear();
                 liberator.modes.show();
                 liberator.statusline.updateTabCount();
-                liberator.buffer.updateBufferList();
                 liberator.tabs.updateSelectionHistory();
 
-                setTimeout(function () { liberator.focusContent(true); }, 10); // just make sure, that no widget has focus
+                if (liberator.options["focuscontent"])
+                    setTimeout(function () { liberator.focusContent(true); }, 10); // just make sure, that no widget has focus
             }, false);
         }
 
@@ -278,7 +286,7 @@ liberator.Events = function () //{{{
             liberator.modes.show();
         }, null);
     }
-    catch (e) { }
+    catch (e) {}
 
 //    getBrowser().addEventListener("submit", function (event)
 //    {
@@ -299,8 +307,8 @@ liberator.Events = function () //{{{
     }
     function exitPopupMode()
     {
-        // gContextMenu is set to NULL by firefox, when a context menu is closed
-        if (!gContextMenu && !activeMenubar)
+        // gContextMenu is set to NULL by Firefox, when a context menu is closed
+        if (typeof gContextMenu != "undefined" && gContextMenu == null && !activeMenubar)
             liberator.modes.remove(liberator.modes.MENU);
     }
     function enterMenuMode()
@@ -386,6 +394,16 @@ liberator.Events = function () //{{{
         return 0;
     }
 
+    function getMacroCompletions(filter)
+    {
+        var macros = [];
+        var tmp = liberator.events.getMacros();
+        for (var item in tmp)
+            macros.push([item, tmp[item]]);
+
+        return [0, liberator.completion.filter(macros, filter)];
+    }
+
     function isFormElemFocused()
     {
         var elt = window.document.commandDispatcher.focusedElement;
@@ -429,14 +447,13 @@ liberator.Events = function () //{{{
 
             // code which should happen for all (also background) newly loaded tabs goes here:
 
-            var url = liberator.buffer.URL;
-            var title = liberator.buffer.title;
+            var url = doc.location.href;
+            var title = doc.title;
 
-            //update history
+            // update history
             if (url && liberator.history)
                 liberator.history.add(url, title);
 
-            liberator.buffer.updateBufferList();
             liberator.autocommands.trigger("PageLoad", url);
 
             // mark the buffer as loaded, we can't use liberator.buffer.loaded
@@ -448,12 +465,19 @@ liberator.Events = function () //{{{
             if (doc == getBrowser().contentDocument)
             {
                 // we want to stay in command mode after a page has loaded
-                // XXX: Does this still causes window map events which is _very_ annoying
-                setTimeout(function () {
-                    var focused = document.commandDispatcher.focusedElement;
-                    if (focused && (typeof focused.value != "undefined") && focused.value.length == 0)
-                        focused.blur();
-                }, 100);
+                // TODO: move somehwere else, as focusing can already happen earlier than on "load"
+                if (liberator.options["focuscontent"])
+                {
+                    setTimeout(function () {
+                        var focused = document.commandDispatcher.focusedElement;
+                        if (focused && (typeof focused.value != "undefined") && focused.value.length == 0)
+                            focused.blur();
+                    }, 100);
+                }
+            }
+            else // background tab
+            {
+                liberator.commandline.echo("Background tab loaded: " + title || url, liberator.commandline.HL_INFOMSG);
             }
         }
     }
@@ -470,7 +494,7 @@ liberator.Events = function () //{{{
         if (liberator.buffer.loaded == 1)
             return true;
 
-        var ms = 15000; // maximum time to wait - TODO: add option
+        var ms = 25000; // maximum time to wait - TODO: add option
         var then = new Date().getTime();
         for (var now = then; now - then < ms; now = new Date().getTime())
         {
@@ -516,12 +540,12 @@ liberator.Events = function () //{{{
 
                 var name = file.leafName.replace(/\.vimp$/i, "");
                 macros[name] = liberator.io.readFile(file).split(/\n/)[0];
-                liberator.log("Macro " + name + " added: " + macros[name], 8);
+                liberator.log("Macro " + name + " added: " + macros[name], 5);
             }
         }
         catch (e)
         {
-            liberator.log("macro directory not found or error reading macro file");
+            liberator.log("Macro directory not found or error reading macro file", 9);
         }
     }, 100);
 
@@ -548,7 +572,7 @@ liberator.Events = function () //{{{
         function () { document.commandDispatcher.rewindFocus(); });
 
     liberator.mappings.add(liberator.modes.all,
-        ["<C-q>"], "Temporarily ignore all " + liberator.config.name + " key bindings",
+        ["<C-z>"], "Temporarily ignore all " + liberator.config.name + " key bindings",
         function () { liberator.modes.passAllKeys = true; });
 
     liberator.mappings.add(liberator.modes.all,
@@ -581,20 +605,23 @@ liberator.Events = function () //{{{
 
     liberator.commands.add(["delmac[ros]"],
         "Delete macros",
-        function (arg)
+        function (args)
         {
-            if (!arg)
+            if (!args)
                 liberator.echoerr("E474: Invalid argument");
             else
-                liberator.events.deleteMacros(arg);
+                liberator.events.deleteMacros(args);
+        },
+        {
+            completer: function (filter) { return getMacroCompletions(filter); }
         });
 
     liberator.commands.add(["macros"],
         "List all macros",
-        function (arg)
+        function (args)
         {
             var str = "<table>";
-            var macroRef = liberator.events.getMacros(arg);
+            var macroRef = liberator.events.getMacros(args);
             for (var item in macroRef)
                str += "<tr><td> " + item + " &nbsp; </td><td>" +
                       liberator.util.escapeHTML(macroRef[item]) + "</td></tr>";
@@ -602,19 +629,22 @@ liberator.Events = function () //{{{
             str += "</table>";
 
             liberator.echo(str, liberator.commandline.FORCE_MULTILINE);
+        },
+        {
+            completer: function (filter) { return getMacroCompletions(filter); }
         });
 
     liberator.commands.add(["pl[ay]"],
         "Replay a recorded macro",
-        function (arg)
+        function (args)
         {
-            if (!arg)
+            if (!args)
                 liberator.echoerr("E474: Invalid argument");
             else
-                liberator.events.playMacro(arg);
+                liberator.events.playMacro(args);
         },
         {
-            completer: function (filter) { return liberator.completion.macros(filter); }
+            completer: function (filter) { return getMacroCompletions(filter); }
         });
 
     /////////////////////////////////////////////////////////////////////////////}}}
@@ -630,7 +660,8 @@ liberator.Events = function () //{{{
             // removeEventListeners() to avoid mem leaks
             window.dump("TODO: remove all eventlisteners\n");
 
-            getBrowser().removeProgressListener(this.progressListener);
+            if (typeof(getBrowser) != "undefined")
+                getBrowser().removeProgressListener(this.progressListener);
 
             window.removeEventListener("popupshown", enterPopupMode, true);
             window.removeEventListener("popuphidden", exitPopupMode, true);
@@ -694,7 +725,8 @@ liberator.Events = function () //{{{
                 try
                 {
                     getWebNavigation().stop(nsIWebNavigation.STOP_ALL);
-                } catch (ex) { }
+                }
+                catch (e) {}
 
                 liberator.buffer.loaded = 1; // even if not a full page load, assume it did load correctly before starting the macro
                 liberator.events.feedkeys(macros[lastMacro], true); // true -> noremap
@@ -922,6 +954,7 @@ liberator.Events = function () //{{{
             if (elem && elem.readOnly)
                 return;
 
+            // liberator.log("onFocusChange: " + elem);
             // dump("=+++++++++=\n" + liberator.util.objectToString(event.target) + "\n")
             // dump (elem + ": " + win + "\n");//" - target: " + event.target + " - origtarget: " + event.originalTarget + " - expltarget: " + event.explicitOriginalTarget + "\n");
 
@@ -959,9 +992,6 @@ liberator.Events = function () //{{{
                     {
                         dump("Compose editor got focus\n");
                         liberator.modes.set(liberator.modes.INSERT, liberator.modes.TEXTAREA);
-                        //setTimeout(function (){ liberator.editor.editWithExternalEditor(); }, 100);
-                        win.blur();
-                        //liberator.editor.editWithExternalEditor();
                     }
                     else if (liberator.mode != liberator.modes.MESSAGE)
                         liberator.mode = liberator.modes.MESSAGE;
@@ -1035,7 +1065,7 @@ liberator.Events = function () //{{{
                         { // a simple if (selection) does not seem to work
                             selection.collapseToStart();
                         }
-                        catch (e) { }
+                        catch (e) {}
                         liberator.commandline.clear();
 
                         liberator.modes.reset();
@@ -1090,7 +1120,7 @@ liberator.Events = function () //{{{
                 if (key == "q") // TODO: should not be hardcoded
                 {
                     liberator.modes.isRecording = false;
-                    liberator.log("Recorded " + currentMacro + ": " + macros[currentMacro], 8);
+                    liberator.log("Recorded " + currentMacro + ": " + macros[currentMacro], 9);
                     event.preventDefault();
                     event.stopPropagation();
                     return true;
@@ -1102,7 +1132,7 @@ liberator.Events = function () //{{{
                 }
             }
 
-            var stop = true; // set to false if we should NOT consume this event but let also firefox handle it
+            var stop = true; // set to false if we should NOT consume this event but let Firefox handle it
 
             var win = document.commandDispatcher.focusedWindow;
             if (win && win.document.designMode == "on" && !liberator.config.isComposeWindow)
@@ -1181,7 +1211,7 @@ liberator.Events = function () //{{{
                     event.stopPropagation();
                     return false;
                 }
-                // if Hit-a-hint mode is on, special handling of keys is required
+                // if Hint mode is on, special handling of keys is required
                 if (liberator.mode == liberator.modes.HINTS)
                 {
                     liberator.hints.onEvent(event);
@@ -1209,7 +1239,6 @@ liberator.Events = function () //{{{
             else
                 map = liberator.mappings.get(liberator.mode, candidateCommand);
 
-
             // counts must be at the start of a complete mapping (10j -> go 10 lines down)
             if (/^[1-9][0-9]*$/.test(liberator.input.buffer + key))
             {
@@ -1314,10 +1343,12 @@ liberator.Events = function () //{{{
                     if (liberator.mode == liberator.modes.COMMAND_LINE)
                     {
                         if (!(liberator.modes.extended & liberator.modes.INPUT_MULTILINE))
-                            stop = !liberator.commandline.onEvent(event); // reroute event in command line mode
+                            liberator.commandline.onEvent(event); // reroute event in command line mode
                     }
-                    else if (liberator.mode != liberator.modes.INSERT)
+                    else if (liberator.mode != liberator.modes.INSERT && liberator.mode != liberator.modes.TEXTAREA)
+                    {
                         liberator.beep();
+                    }
                 }
             }
 
@@ -1368,8 +1399,16 @@ liberator.Events = function () //{{{
                     {
                         liberator.buffer.loaded = 0;
                         liberator.statusline.updateProgress(0);
-                        setTimeout (function () { liberator.modes.reset(false); },
-                            liberator.mode == liberator.modes.HINTS ? 500 : 0);
+
+                        liberator.autocommands.trigger("PageLoadPre", liberator.buffer.URL);
+
+                        // don't reset mode if a frame of the frameset gets reloaded which
+                        // is not the focused frame
+                        if (document.commandDispatcher.focusedWindow == webProgress.DOMWindow)
+                        {
+                            setTimeout (function () { liberator.modes.reset(false); },
+                                liberator.mode == liberator.modes.HINTS ? 500 : 0);
+                        }
                     }
                     else if (flags & Components.interfaces.nsIWebProgressListener.STATE_STOP)
                     {
@@ -1403,6 +1442,8 @@ liberator.Events = function () //{{{
                 liberator.statusline.updateUrl();
                 liberator.statusline.updateProgress();
 
+                liberator.autocommands.trigger("LocationChange", liberator.buffer.URL);
+
                 // if this is not delayed we get the position of the old buffer
                 setTimeout(function () { liberator.statusline.updateBufferPosition(); }, 100);
             },
@@ -1471,7 +1512,7 @@ liberator.Events = function () //{{{
                 }
              }
         }
-    };
+    }; //}}}
 
     window.XULBrowserWindow = eventManager.progressListener;
     window.QueryInterface(Components.interfaces.nsIInterfaceRequestor)
@@ -1484,7 +1525,7 @@ liberator.Events = function () //{{{
     {
         getBrowser().addProgressListener(eventManager.progressListener, Components.interfaces.nsIWebProgress.NOTIFY_ALL);
     }
-    catch(e) { }
+    catch (e) {}
 
     eventManager.prefObserver.register();
 
@@ -1493,7 +1534,7 @@ liberator.Events = function () //{{{
     window.addEventListener("keyup",    eventManager.onKeyUpOrDown, true);
 
     return eventManager;
-    //}}}
+
 }; //}}}
 
 // vim: set fdm=marker sw=4 ts=4 et:
