@@ -41,6 +41,34 @@ liberator.IO = function () //{{{
     var extname = liberator.config.name.toLowerCase(); // "vimperator" or "muttator"
     var lastRunCommand = ""; // updated whenever the users runs a command with :!
 
+    /////////////////////////////////////////////////////////////////////////////}}}
+    ////////////////////// OPTIONS ////////////////////////////////////////////////
+    /////////////////////////////////////////////////////////////////////////////{{{
+
+    var shell, shellcmdflag;
+
+    if (WINDOWS)
+    {
+        shell = "cmd.exe";
+        // TODO: does setting 'shell' to "something containing sh"
+        // automatically update 'shellcmdflag' on Windows in Vim?
+        shellcmdflag = "/c";
+    }
+    else
+    {
+        shell = environmentService.get("SHELL") || "sh";
+        shellcmdflag = "-c";
+    }
+
+    // TODO: setter should expand environment variables
+    liberator.options.add(["shell", "sh"],
+        "Shell to use for executing :! and :run commands",
+        "string", shell);
+
+    liberator.options.add(["shellcmdflag", "shcf"],
+        "Flag passed to shell when executing :! and :run commands",
+        "string", shellcmdflag);
+
     /////////////////////////////////////////////////////////////////////////////}}}
     ////////////////////// COMMANDS ////////////////////////////////////////////////
     /////////////////////////////////////////////////////////////////////////////{{{
@@ -61,27 +89,27 @@ liberator.IO = function () //{{{
 
     liberator.commands.add(["pw[d]"],
         "Print the current directory name",
-        function (args)
+        function ()
         {
-            if (args)
-                liberator.echoerr("E488: Trailing characters");
-            else
-                liberator.echo(liberator.io.getCurrentDirectory());
-        });
+            liberator.echo(liberator.io.getCurrentDirectory());
+        },
+        { argCount: "0" });
 
     // mkv[imperatorrc] or mkm[uttatorrc]
-    liberator.commands.add(["mk" + extname.substr(0,1) + "[" + extname.substr(1) + "rc]"],
+    liberator.commands.add(["mk" + extname.substr(0, 1) + "[" + extname.substr(1) + "rc]"],
         "Write current key mappings and changed options to the config file",
         function (args, special)
         {
             // TODO: "E172: Only one file name allowed"
             var filename;
             if (args)
+            {
                 filename = args;
+            }
             else
             {
                 filename = (navigator.platform == "Win32") ? "~/_" : "~/.";
-                filename += extname+ "rc";
+                filename += extname + "rc";
             }
 
             var file = liberator.io.getFile(filename);
@@ -338,7 +366,10 @@ liberator.IO = function () //{{{
             var file = Components.classes["@mozilla.org/file/local;1"].
                                   createInstance(Components.interfaces.nsILocalFile);
 
-            var tmpname = liberator.config.name.toLowerCase() + ".tmp"
+            var tmpname = liberator.config.name.toLowerCase() + ".tmp";
+            if (liberator.config.name == "Muttator")
+                tmpname = "mutt-ator-mail"; // to allow vim to :set ft=mail automatically
+
             if (WINDOWS)
             {
                 var dir = environmentService.get("TMP") || environmentService.get("TEMP") || "C:\\";
@@ -484,13 +515,13 @@ lookup:
                             }
                         }
                     }
-                    catch (e) { }
+                    catch (e) {}
                 }
             }
 
             if (!file.exists())
             {
-                liberator.echoerr("command not found: " + program);
+                liberator.echoerr("Command not found: " + program);
                 return -1;
             }
 
@@ -504,7 +535,6 @@ lookup:
 
         // when https://bugzilla.mozilla.org/show_bug.cgi?id=68702 is fixed
         // is fixed, should use that instead of a tmpfile
-        // TODO: add shell/shellcmdflag options to replace "sh" and "-c"
         system: function (str, input)
         {
             var fileout = ioManager.createTempFile();
@@ -524,11 +554,8 @@ lookup:
                 command += " < \"" + filein.path.replace('"', '\\"') + "\"";
             }
 
-            var res;
-            if (WINDOWS)
-                res = ioManager.run("cmd.exe", ["/C", command], true);
-            else
-                res = ioManager.run("sh", ["-c", command], true);
+            // FIXME: why is this return value being ignored?
+            var res = ioManager.run(liberator.options["shell"], [liberator.options["shellcmdflag"], command], true);
 
             var output = ioManager.readFile(fileout);
             fileout.remove(false);
@@ -617,9 +644,10 @@ lookup:
                     liberator.echoerr(e);
             }
         }
-    };
+    }; //}}}
+
     return ioManager;
-    //}}}
+
 }; //}}}
 
 // vim: set fdm=marker sw=4 ts=4 et:
