@@ -31,14 +31,21 @@ const liberator = (function () //{{{
     ////////////////////////////////////////////////////////////////////////////////
     ////////////////////// PRIVATE SECTION /////////////////////////////////////////
     /////////////////////////////////////////////////////////////////////////////{{{
+
     var callbacks = [];
 
+    function loadModule(name, func)
+    {
+        liberator.log("Loading module " + name + "...", 0);
+        liberator[name] = func();
+    }
+
     // Only general options are added here, which are valid for all vimperator like extensions
     function addOptions()
     {
         liberator.options.add(["guioptions", "go"],
             "Show or hide certain GUI elements like the menu or toolbar",
-            "charlist", "",
+            "charlist", liberator.config.defaults.guioptions || "",
             {
                 setter: function (value)
                 {
@@ -51,7 +58,7 @@ const liberator = (function () //{{{
                             {
                                 document.getElementById(elem).collapsed = (value.indexOf(option.toString()) < 0);
                             }
-                            catch (e) { }
+                            catch (e) {}
                         });
                     }
                 },
@@ -65,28 +72,6 @@ const liberator = (function () //{{{
                 }
             });
 
-        liberator.options.add(["titlestring"], // TODO: broken for Thunderbird
-            "Change the title of the window",
-            "string", "Vimperator",
-            {
-                setter: function (value)
-                {
-                    try
-                    {
-                        var id = liberator.config.mainWindowID || "main-window";
-                        document.getElementById(id).setAttribute("titlemodifier", value);
-                        if (window.content.document.title.length > 0)
-                            document.title = window.content.document.title + " - " + value;
-                        else
-                            document.title = value;
-                    }
-                    catch (e)
-                    {
-                        liberator.log("Couldn't set titlestring", 1);
-                    }
-                },
-            });
-
         liberator.options.add(["verbose", "vbs"],
             "Define which type of messages are logged",
             "number", 0,
@@ -98,7 +83,7 @@ const liberator = (function () //{{{
             "Use visual bell instead of beeping on errors",
             "boolean", false,
             {
-                setter: function (value) { liberator.options.setPref("accessibility.typeaheadfind.enablesound", !value); },
+                setter: function (value) { liberator.options.setPref("accessibility.typeaheadfind.enablesound", !value); }
             });
 
         liberator.options.add(["visualbellstyle", "t_vb"],
@@ -132,14 +117,19 @@ const liberator = (function () //{{{
             function ()
             {
                 liberator.open("chrome://mozapps/content/extensions/extensions.xul",
-                    (liberator.options.newtab &&
-                        (liberator.options.newtab == "all" || liberator.options.newtab.split(",").indexOf("addons") != -1)) ?
+                    (liberator.options["newtab"] &&
+                        (liberator.options["newtab"] == "all" || liberator.options["newtab"].split(",").indexOf("addons") != -1)) ?
                             liberator.NEW_TAB: liberator.CURRENT_TAB);
-            });
+            },
+            { argCount: "0" });
 
         liberator.commands.add(["beep"],
             "Play a system beep",
-            function () { liberator.beep(); });
+            function ()
+            {
+                liberator.beep();
+            },
+            { argCount: "0" });
 
         liberator.commands.add(["dia[log]"],
             "Open a " + liberator.config.appName + " dialog",
@@ -161,7 +151,70 @@ const liberator = (function () //{{{
                 }
             },
             {
-                completer: function (filter) { return liberator.completion.dialog(filter); }
+                completer: function (filter)
+                {
+                    return [0, liberator.completion.filter(liberator.config.dialogs || [], filter)];
+                }
+            });
+
+        // TODO: move this
+        function getMenuItems()
+        {
+            var menubar = document.getElementById(liberator.config.guioptions["m"]);
+            var items = [];
+
+            for (var i = 0; i < menubar.childNodes.length; i++)
+            {
+                (function (item, path)
+                {
+                    if (item.childNodes.length == 0 && item.localName == "menuitem"
+                        && !/rdf:http:/.test(item.label)) // FIXME
+                    {
+                        item.fullMenuPath = path += item.label;
+                        items.push(item);
+                    }
+                    else
+                    {
+                        if (item.localName == "menu")
+                            path += item.label + "."; 
+
+                        for (var j = 0; j < item.childNodes.length; j++)
+                            arguments.callee(item.childNodes[j], path);
+                    }
+                })(menubar.childNodes[i], "");
+            }
+
+            return items;
+        }
+
+        liberator.commands.addUserCommand(["em[enu]"],
+            "Execute the specified menu item from the command line",
+            function (args)
+            {
+                var item = args.string;
+                var items = getMenuItems();
+
+                if (!items.some(function (i) { return i.fullMenuPath == item; }))
+                {
+                    liberator.echoerr("E334: Menu not found: " + item);
+                    return;
+                }
+
+                for (var i = 0; i < items.length; i++)
+                {
+                    if (items[i].fullMenuPath == item)
+                        items[i].doCommand(); 
+                }
+            },
+            {
+                argCount: "+", // NOTE: single arg may contain unescaped whitespace
+                completer: function (filter)
+                {
+                    var completions = getMenuItems().map(function (item) {
+                        return [item.fullMenuPath, item.label];
+                    });
+                    return [0, liberator.completion.filter(completions, filter)];	
+                }
             });
 
         liberator.commands.add(["exe[cute]"],
@@ -182,13 +235,14 @@ const liberator = (function () //{{{
                 usage += "</table>";
 
                 liberator.echo(usage, liberator.commandline.FORCE_MULTILINE);
-            });
+            },
+            { argCount: "0" });
 
         liberator.commands.add(["h[elp]"],
             "Display help",
             function (args) { liberator.help(args); },
             {
-                completer: function (filter) { return liberator.completion.help(filter); }
+                completer: function (filter) { return getHelpCompletions(filter); }
             });
 
         liberator.commands.add(["javas[cript]", "js"],
@@ -198,8 +252,8 @@ const liberator = (function () //{{{
                 if (special) // open javascript console
                 {
                     liberator.open("chrome://global/content/console.xul",
-                        (liberator.options.newtab &&
-                            (liberator.options.newtab == "all" || liberator.options.newtab.split(",").indexOf("javascript") != -1)) ?
+                        (liberator.options["newtab"] &&
+                            (liberator.options["newtab"] == "all" || liberator.options["newtab"].split(",").indexOf("javascript") != -1)) ?
                                 liberator.NEW_TAB : liberator.CURRENT_TAB);
                 }
                 else
@@ -225,7 +279,7 @@ const liberator = (function () //{{{
                     {
                         try
                         {
-                            eval("with(liberator){" + args + "}");
+                            eval("with(liberator) {" + args + "}");
                         }
                         catch (e)
                         {
@@ -242,14 +296,9 @@ const liberator = (function () //{{{
             "Execute Normal mode commands",
             function (args, special)
             {
-                if (!args)
-                {
-                    liberator.echoerr("E471: Argument required");
-                    return;
-                }
-
-                liberator.events.feedkeys(args, special);
-            });
+                liberator.events.feedkeys(args.string, special);
+            },
+            { argCount: "+" });
 
         liberator.commands.add(["q[uit]"],
             liberator.has("tabs") ? "Quit current tab" : "Quit application",
@@ -259,11 +308,16 @@ const liberator = (function () //{{{
                     liberator.tabs.remove(getBrowser().mCurrentTab, 1, false, 1);
                 else
                     liberator.quit(false);
-            });
+            },
+            { argCount: "0" });
 
         liberator.commands.add(["res[tart]"],
             "Force " + liberator.config.appName + " to restart",
-            function () { liberator.restart(); });
+            function ()
+            {
+                liberator.restart();
+            },
+            { argCount: "0" });
 
         liberator.commands.add(["time"],
             "Profile a piece of code or run a command multiple times",
@@ -284,7 +338,7 @@ const liberator = (function () //{{{
                         else
                         {
                             while (i--)
-                                eval("with(liberator){" + args + "}");
+                                eval("with(liberator) {" + args + "}");
                         }
 
                         if (special)
@@ -347,6 +401,15 @@ const liberator = (function () //{{{
                 {
                     liberator.echoerr(e);
                 }
+            },
+            {
+                completer: function (filter)
+                {
+                    if (/^:/.test(filter))
+                        return liberator.completion.ex(filter);
+                    else
+                        return liberator.completion.javascript(filter);
+                }
             });
 
         liberator.commands.add(["ve[rsion]"],
@@ -359,7 +422,8 @@ const liberator = (function () //{{{
                     liberator.echo(":" + liberator.util.escapeHTML(liberator.commandline.getCommand()) + "\n" +
                                     liberator.config.name + " " + liberator.version +
                                     " running on:\n" + navigator.userAgent);
-            });
+            },
+            { argCount: "0" });
 
         liberator.commands.add(["viu[sage]"],
             "List all mappings with a short description",
@@ -375,7 +439,35 @@ const liberator = (function () //{{{
                 usage += "</table>";
 
                 liberator.echo(usage, liberator.commandline.FORCE_MULTILINE);
-            });
+            },
+            { argCount: "0" });
+    }
+
+    function getHelpCompletions(filter)
+    {
+        var files = liberator.config.helpFiles || [];
+        var res = [];
+
+        for (var i = 0; i < files.length; i++)
+        {
+            try
+            {
+                var xmlhttp = new XMLHttpRequest();
+                xmlhttp.open("GET", "chrome://" + liberator.config.name.toLowerCase() + "/locale/" + files[i], false);
+                xmlhttp.send(null);
+            }
+            catch (e)
+            {
+                liberator.log("Error opening chrome://" + liberator.config.name.toLowerCase() + "/locale/" + files[i], 1);
+                continue;
+            }
+            var doc = xmlhttp.responseXML;
+            var elems = doc.getElementsByClassName("tag");
+            for (var j = 0; j < elems.length; j++)
+                res.push([elems[j].textContent, files[i]]);
+        }
+
+        return [0, liberator.completion.filter(res, filter)];
     }
 
     // initially hide all GUI, it is later restored unless the user has :set go= or something
@@ -391,7 +483,7 @@ const liberator = (function () //{{{
                 {
                     document.getElementById(elem).collapsed = true;
                 }
-                catch (e) { }
+                catch (e) {}
             });
         }
     }
@@ -413,8 +505,8 @@ const liberator = (function () //{{{
 
         forceNewTab: false,
 
-        // 1.1 and 2008/06/03 19:37:22 are replaced by the Makefile
-        version: "1.1 (created: 2008/06/03 19:37:22)",
+        // 1.2 and 2008/08/16 18:51:52 are replaced by the Makefile
+        version: "1.2 (created: 2008/08/16 18:51:52)",
 
         // TODO: move to events.js?
         input: {
@@ -430,6 +522,7 @@ const liberator = (function () //{{{
         //  "cancel"
         //  "complete"
         //  TODO: "zoom": if the zoom value of the current buffer changed
+        //  TODO: move to ui.js?
         registerCallback: function (type, mode, func)
         {
             // TODO: check if callback is already registered
@@ -457,7 +550,7 @@ const liberator = (function () //{{{
                 {
                     // flash the visual bell
                     var popup = document.getElementById("liberator-visualbell");
-                    var win = getBrowser().mPanelContainer;
+                    var win = liberator.config.visualbellWindow;
                     var box = document.getBoxObjectFor(win);
 
                     popup.style.cssText = vbs;
@@ -473,6 +566,7 @@ const liberator = (function () //{{{
                                    getService(Components.interfaces.nsISound);
                 soundService.beep();
             }
+            return false; // so you can do: if (...) return liberator.beep();
         },
 
         // Execute an ex command like str=":zoom 300"
@@ -528,7 +622,7 @@ const liberator = (function () //{{{
                     gDBView.selection.select(i);
                 }
             }
-            catch (e) { }
+            catch (e) {}
 
             var elem = liberator.config.mainWidget || window.content;
             if (elem && (elem != document.commandDispatcher.focusedElement))
@@ -536,6 +630,8 @@ const liberator = (function () //{{{
         },
 
         // partial sixth level expression evaluation
+        // TODO: what is that really needed for, and where could it be used?
+        //       Or should it be removed? (c) Viktor
         eval: function (string)
         {
             string = string.toString().replace(/^\s*/, "").replace(/\s*$/, "");
@@ -608,6 +704,10 @@ const liberator = (function () //{{{
             }
         },
 
+        echo:    function (str, flags) { liberator.commandline.echo(str, liberator.commandline.HL_NORMAL, flags); },
+
+        echoerr: function (str, flags) { liberator.commandline.echo(str, liberator.commandline.HL_ERRORMSG, flags); },
+
         // return true, if this VIM-like extension has a certain feature
         has: function (feature)
         {
@@ -617,12 +717,14 @@ const liberator = (function () //{{{
 
         help: function (topic)
         {
-            var where = (liberator.options.newtab && (liberator.options.newtab == "all" || liberator.options.newtab.split(",").indexOf("help") != -1)) ?
-                        liberator.NEW_TAB : liberator.CURRENT_TAB;
+            var where = (liberator.options["newtab"] && (liberator.options["newtab"] == "all" ||
+                         liberator.options["newtab"].split(",").indexOf("help") != -1)) ?
+                            liberator.NEW_TAB : liberator.CURRENT_TAB;
 
             function jumpToTag(file, tag)
             {
                 liberator.open("chrome://" + liberator.config.name.toLowerCase() + "/locale/" + file, where);
+                // TODO: it would be better wo wait for pageLoad
                 setTimeout(function () {
                     var elem = liberator.buffer.getElement('@class="tag" and text()="' + tag + '"');
                     if (elem)
@@ -638,7 +740,7 @@ const liberator = (function () //{{{
                 return;
             }
 
-            var [, items] = liberator.completion.help(topic);
+            var [, items] = getHelpCompletions(topic);
             var partialMatch = -1;
             for (var i = 0; i < items.length; i++)
             {
@@ -659,11 +761,25 @@ const liberator = (function () //{{{
                 liberator.echoerr("E149: Sorry, no help for " + topic);
         },
 
+        globalVariables: {},
+
+        loadModule: function (name, func) { loadModule(name, func); },
+
         // logs a message to the javascript error console
         // if msg is an object, it is beautified
         log: function (msg, level)
         {
-            //if (liberator.options.getPref("verbose") >= level) // FIXME: hangs liberator, probably timing issue --mst
+            var verbose = 0;
+            if (typeof level != "number")
+                level = 1;
+
+            // liberator.options does not exist at the very beginning
+            if (liberator.options)
+                verbose = liberator.options["verbose"];
+
+            if (level > verbose)
+                return;
+
             if (typeof msg == "object")
                 msg = liberator.util.objectToString(msg, false);
 
@@ -746,10 +862,16 @@ const liberator = (function () //{{{
             return true;
         },
 
+        // namespace for plugins/scripts. Actually (only) the active plugin must/can set a
+        // v.plugins.mode = <str> string to show on v.modes.CUSTOM
+        // v.plugins.stop = <func> hooked on a v.modes.reset()
+        // v.plugins.onEvent = <func> function triggered, on keypresses (unless <esc>) (see events.js)
+        plugins: {},
+
         // quit liberator, no matter how many tabs/windows are open
         quit: function (saveSession)
         {
-            liberator.autocommands.trigger("BrowserExit", "");
+            liberator.autocommands.trigger("Quit", "");
 
             if (saveSession)
                 liberator.options.setPref("browser.startup.page", 3); // start with saved session
@@ -761,8 +883,6 @@ const liberator = (function () //{{{
 
         restart: function ()
         {
-            liberator.autocommands.trigger("BrowserRestart", "");
-
             const nsIAppStartup = Components.interfaces.nsIAppStartup;
 
             // notify all windows that an application quit has been requested.
@@ -776,6 +896,8 @@ const liberator = (function () //{{{
             if (cancelQuit.data)
                 return;
 
+            liberator.autocommands.trigger("Quit", "");
+
             // notify all windows that an application quit has been granted.
             os.notifyObservers(null, "quit-application-granted", null);
 
@@ -793,63 +915,35 @@ const liberator = (function () //{{{
                 .quit(nsIAppStartup.eRestart | nsIAppStartup.eAttemptQuit);
         },
 
+
         // TODO: move to {muttator,vimperator,...}.js
         // this function is called, when the chrome is ready
         startup: function ()
         {
-            function log(module) { liberator.log("Loading module " + module + "...", 3); };
+            liberator.log("Initializing liberator object...", 0);
 
-            liberator.log("Initializing vimperator object...", 1);
             // commands must always be the first module to be initialized
-            log("commands");       liberator.commands = liberator.Commands(); addCommands();
-            log("options");        liberator.options  = liberator.Options();  addOptions();
-            log("mappings");       liberator.mappings = liberator.Mappings(); addMappings();
-            log("events");         liberator.events   = liberator.Events();
-            log("commandline");    liberator.commandline   = liberator.CommandLine();
-            log("search");         liberator.search        = liberator.Search();
-            log("preview window"); liberator.previewwindow = liberator.InformationList("liberator-previewwindow", { incrementalFill: false, maxItems: 10 });
-            log("buffer window");  liberator.bufferwindow  = liberator.InformationList("liberator-bufferwindow", { incrementalFill: false, maxItems: 10 });
-            log("statusline");     liberator.statusline    = liberator.StatusLine();
-            log("buffer");         liberator.buffer        = liberator.Buffer();
-            log("editor");         liberator.editor        = liberator.Editor();
-            log("autocommands");   liberator.autocommands  = liberator.AutoCommands();
-            log("io");             liberator.io            = liberator.IO();
-            log("completion");     liberator.completion    = liberator.Completion();
-
-            // optional modules
-            if (liberator.has("bookmarks"))  { log("bookmarks");  liberator.bookmarks  = liberator.Bookmarks(); }
-            if (liberator.has("history"))    { log("history");    liberator.history    = liberator.History(); }
-            if (liberator.has("mail") && liberator.Mail)       { log("mail");       liberator.mail       = liberator.Mail(); }
-            if (liberator.has("tabs") && liberator.Tabs)       { log("tabs");       liberator.tabs       = liberator.Tabs(); }
-            if (liberator.has("marks"))      { log("marks");      liberator.marks      = liberator.Marks(); }
-            if (liberator.has("quickmarks")) { log("quickmarks"); liberator.quickmarks = liberator.QuickMarks(); }
-            if (liberator.has("hints"))      { log("hints");      liberator.hints      = liberator.Hints(); }
-            if (liberator.has("addressbook") && liberator.Addressbook)      { log("addressbook");      liberator.addressbook      = liberator.Addressbook(); }
-
-            liberator.log("All modules loaded", 3);
+            loadModule("commands",     liberator.Commands); addCommands();
+            loadModule("options",      liberator.Options);  addOptions();
+            loadModule("mappings",     liberator.Mappings); addMappings();
+            loadModule("events",       liberator.Events);
+            loadModule("commandline",  liberator.CommandLine);
+            loadModule("statusline",   liberator.StatusLine);
+            loadModule("buffer",       liberator.Buffer);
+            loadModule("editor",       liberator.Editor);
+            loadModule("autocommands", liberator.AutoCommands);
+            loadModule("io",           liberator.IO);
+            loadModule("completion",   liberator.Completion);
 
             // This adds options/mappings/commands which are only valid in this particular extension
             if (liberator.config.init)
-            {
                 liberator.config.init();
-                // liberator.log("Loaded additional mappings, etc. for " + liberator.config.name, 3);
-            }
 
-            // we define some shortcuts to functions which are used often
-            liberator.echo    = function (str, flags) { liberator.commandline.echo(str, liberator.commandline.HL_NORMAL, flags); };
-            liberator.echoerr = function (str, flags) { liberator.commandline.echo(str, liberator.commandline.HL_ERRORMSG, flags); };
-
-            liberator.globalVariables = {};
-
-            // namespace for plugins/scripts. Actually (only) the active plugin must/can set a
-            // v.plugins.mode = <str> string to show on v.modes.CUSTOM
-            // v.plugins.stop = <func> hooked on a v.modes.reset()
-            // v.plugins.onEvent = <func> function triggered, on keypresses (unless <esc>) (see events.js)
-            liberator.plugins = {};
+            liberator.log("All modules loaded", 3);
 
             // TODO: move elsewhere
             liberator.registerCallback("submit", liberator.modes.EX, function (command) { liberator.execute(command); });
-            liberator.registerCallback("complete", liberator.modes.EX, function (str) { return liberator.completion.exTabCompletion(str); });
+            liberator.registerCallback("complete", liberator.modes.EX, function (str) { return liberator.completion.ex(str); });
 
             // first time intro message
             if (liberator.options.getPref("extensions." + liberator.config.name.toLowerCase() + ".firsttime", true))
@@ -908,10 +1002,13 @@ const liberator = (function () //{{{
                     if (option.setter && !option.hasChanged)
                         option.reset();
                 }
+
+                liberator.autocommands.trigger("Startup", "");
             }, 0);
 
             liberator.statusline.update();
-            liberator.log(liberator.config.name + " fully initialized", 1);
+
+            liberator.log(liberator.config.name + " fully initialized", 0);
         },
 
         shutdown: function ()
@@ -979,7 +1076,8 @@ const liberator = (function () //{{{
 			thread.dispatch(new CallbackEvent(func, args), thread.DISPATCH_SYNC);
         }
 
-    }; //}}}
+    };
+    //}}}
 })(); //}}}
 
 // called when the chrome is fully loaded and before the main window is shown
