@@ -38,13 +38,13 @@ liberator.Mail = function () //{{{
     var selectMessageReverse = false;
 
     var folderListener = {
-        OnItemAdded: function (parentItem, item) { },
-        OnItemRemoved: function (parentItem, item) { },
-        OnItemPropertyChanged: function (item, property, oldValue, newValue) { },
-        OnItemIntPropertyChanged: function (item, property, oldValue, newValue) { },
-        OnItemBoolPropertyChanged: function (item, property, oldValue, newValue) { },
-        OnItemUnicharPropertyChanged: function (item, property, oldValue, newValue) { },
-        OnItemPropertyFlagChanged: function (item, property, oldFlag, newFlag) { },
+        OnItemAdded: function (parentItem, item) {},
+        OnItemRemoved: function (parentItem, item) {},
+        OnItemPropertyChanged: function (item, property, oldValue, newValue) {},
+        OnItemIntPropertyChanged: function (item, property, oldValue, newValue) {},
+        OnItemBoolPropertyChanged: function (item, property, oldValue, newValue) {},
+        OnItemUnicharPropertyChanged: function (item, property, oldValue, newValue) {},
+        OnItemPropertyFlagChanged: function (item, property, oldFlag, newFlag) {},
 
         OnItemEvent: function (folder, event)
         {
@@ -54,7 +54,7 @@ liberator.Mail = function () //{{{
                 if (folder)
                 {
                     var msgFolder = folder.QueryInterface(Components.interfaces.nsIMsgFolder);
-                    dump (msgFolder.prettiestName + " loaded\n");
+                    liberator.autocommands.trigger("FolderLoaded", msgFolder);
 
                     // Jump to a message when requested
                     var indices = [];
@@ -73,13 +73,13 @@ liberator.Mail = function () //{{{
                     }
                 }
             }
-            /*else if (eventType == "ImapHdrDownloaded") { }
-            else if (eventType == "DeleteOrMoveMsgCompleted") { }
-            else if (eventType == "DeleteOrMoveMsgFailed") { }
-            else if (eventType == "AboutToCompact") { }
-            else if (eventType == "CompactCompleted") { }
-            else if (eventType == "RenameCompleted") { }
-            else if (eventType == "JunkStatusChanged") { }*/
+            /*else if (eventType == "ImapHdrDownloaded") {}
+            else if (eventType == "DeleteOrMoveMsgCompleted") {}
+            else if (eventType == "DeleteOrMoveMsgFailed") {}
+            else if (eventType == "AboutToCompact") {}
+            else if (eventType == "CompactCompleted") {}
+            else if (eventType == "RenameCompleted") {}
+            else if (eventType == "JunkStatusChanged") {}*/
         }
     };
 
@@ -89,6 +89,25 @@ liberator.Mail = function () //{{{
     var notifyFlags = nsIFolderListener.intPropertyChanged | nsIFolderListener.event;
     mailSession.AddFolderListener(folderListener, notifyFlags);
 
+    function getFolderCompletions(filter)
+    {
+        var completions = [];
+        var folders = liberator.mail.getFolders();
+        for (var folder in folders)
+        {
+            completions.push([folders[folder].server.prettyName + ": "
+                              + folders[folder].name,
+                             "Unread: " + folders[folder].getNumUnread(false)]);
+        }
+
+        return [0, liberator.completion.filter(completions, filter)];
+    }
+
+    function getRSSUrl()
+    {
+        return gDBView.hdrForFirstSelectedMessage.messageId.replace(/(#.*)?@.*$/, "");
+    }
+
     function moveOrCopy(copy, destinationFolder, operateOnThread)
     {
         if (!destinationFolder)
@@ -118,12 +137,12 @@ liberator.Mail = function () //{{{
 
         if (copy)
         {
-            MsgCopyMessage(folders[0].URI);
+            MsgCopyMessage(folders[0]);
             setTimeout(function () { liberator.echo(count + " message(s) copied to " + folders[0].prettyName); }, 100);
         }
         else
         {
-            MsgMoveMessage(folders[0].URI);
+            MsgMoveMessage(folders[0]);
             setTimeout(function () { liberator.echo(count + " message(s) moved to " + folders[0].prettyName); }, 100);
         }
         return true;
@@ -228,20 +247,21 @@ liberator.Mail = function () //{{{
 	    function ()
 	    {
 	    	if (gDBView && gDBView.selection.count < 1)
-	    	{
-	    		liberator.echoerr("Message is not selected");
-	    		return;
-	    	}
+	    		return liberator.beep();
+
 	    	MsgOpenNewTabForMessage();
 	    });
 
+    /*liberator.mappings.add([liberator.modes.NORMAL],
+        ["o"], "Open a message",
+        function () { liberator.commandline.open(":", "open ", liberator.modes.EX); });*/
+
     liberator.mappings.add(modes, ["<Space>"],
         "Scroll message or select next unread one",
         function () { return true; },
         { flags: liberator.Mappings.flags.ALLOW_EVENT_ROUTING });
 
-    // TODO: change to "t" probably
-    liberator.mappings.add(modes, ["x"],
+    liberator.mappings.add(modes, ["t"],
         "Select thread",
         function () { gDBView.ExpandAndSelectThreadByIndex(GetThreadTree().currentIndex, false); });
 
@@ -327,6 +347,10 @@ liberator.Mail = function () //{{{
         "Reply to sender",
         function () { goDoCommand("cmd_reply"); });
 
+    liberator.mappings.add(modes, ["R"],
+        "Reply to all",
+        function () { goDoCommand("cmd_replyall"); });
+
     liberator.mappings.add(modes, ["f"],
         "Forward message",
         function () { goDoCommand("cmd_forward"); });
@@ -525,49 +549,36 @@ liberator.Mail = function () //{{{
         { flags: liberator.Mappings.flags.COUNT });
 
     // tagging messages
-    liberator.mappings.add(modes, ["tr"],
-        "Toggle selected messages read",
-        function ()
+    liberator.mappings.add(modes, ["l"],
+        "Label message",
+        function (arg)
         {
             if (!GetSelectedMessages())
+                return liberator.beep();
+
+            switch (arg)
             {
-                liberator.beep();
-                return;
+                case "r": MsgMarkMsgAsRead(); break;
+                case "s": MsgMarkMsgAsFlagged(); break;
+                case "i": ToggleMessageTagKey(1); break; // Important
+                case "w": ToggleMessageTagKey(2); break; // Work
+                case "p": ToggleMessageTagKey(3); break; // Personal
+                case "t": ToggleMessageTagKey(4); break; // TODO
+                case "l": ToggleMessageTagKey(5); break; // Later
+                default:  liberator.beep();
             }
-
-            MsgMarkMsgAsRead();
-        });
-    liberator.mappings.add(modes, ["tR"],
-        "Tag thread as read",
-        function ()
+        },
         {
-            // TODO: ensure thread or beep
-
-            MsgMarkThreadAsRead();
-        });
-
-    liberator.mappings.add(modes, ["ts"],
-        "Toggle selected messages starred",
-        function ()
-        {
-            if (!GetSelectedMessages())
-            {
-                liberator.beep();
-                return;
-            }
-
-            MsgMarkMsgAsFlagged();
+            flags: liberator.Mappings.flags.ARGUMENT
         });
 
+    // TODO: change binding?
     liberator.mappings.add(modes, ["T"],
         "Mark current folder as read",
         function ()
         {
             if (liberator.mail.currentFolder.isServer)
-            {
-                liberator.beep();
-                return;
-            }
+                return liberator.beep();
 
             liberator.mail.currentFolder.markAllMessagesRead();
         });
@@ -581,6 +592,65 @@ liberator.Mail = function () //{{{
             });
         });
 
+    // DISPLAY OPTIONS
+    liberator.mappings.add(modes, ["h"],
+        "Toggle displayed headers",
+        function ()
+        {
+            var value = gPrefBranch.getIntPref("mail.show_headers", 2);
+            gPrefBranch.setIntPref("mail.show_headers", value == 2 ? 1 : 2);
+            MsgReload();
+        });
+
+    liberator.mappings.add(modes, ["x"],
+        "Toggle HTML message display",
+        function ()
+        {
+            var want_html = (gPrefBranch.getIntPref("mailnews.display.html_as", 1) == 1);
+            liberator.mail.setHTML(want_html ? 1 : 0);
+        });
+
+    // YANKING TEXT
+    liberator.mappings.add(modes, ["Y"],
+        "Yank subject",
+        function ()
+        {
+            try
+            {
+                var subject = gDBView.hdrForFirstSelectedMessage.mime2DecodedSubject;
+                liberator.util.copyToClipboard(subject, true);
+            }
+            catch (e) { liberator.beep(); }
+        });
+
+    liberator.mappings.add(modes, ["y"],
+        "Yank sender or feed URL",
+        function ()
+        {
+            try
+            {
+                if (liberator.mail.currentAccount.server.type == "rss")
+                    liberator.util.copyToClipboard(getRSSUrl(), true);
+                else
+                    liberator.util.copyToClipboard(gDBView.hdrForFirstSelectedMessage.mime2DecodedAuthor, true);
+            }
+            catch (e) { liberator.beep(); }
+        });
+
+    // RSS specific mappings
+    liberator.mappings.add(modes, ["p"],
+        "Open RSS message in browser",
+        function ()
+        {
+            try
+            {
+                if (liberator.mail.currentAccount.server.type == "rss")
+                    messenger.launchExternalURL(getRSSUrl());
+                // TODO: what to do for non-rss message?
+            }
+            catch (e) { liberator.beep(); }
+        });
+
     /////////////////////////////////////////////////////////////////////////////}}}
     ////////////////////// COMMANDS ////////////////////////////////////////////////
     /////////////////////////////////////////////////////////////////////////////{{{
@@ -601,93 +671,73 @@ liberator.Mail = function () //{{{
                 SelectFolder(folder.URI);
         },
         {
-            completer: function (filter) { return liberator.completion.mail(filter); }
+            completer: function (filter) { return getFolderCompletions(filter); }
         });
 
     liberator.commands.add(["m[essage]"],
         "Write a new message",
         function (args, special, count)
         {
-            var res = liberator.commands.parseArgs(args, this.args);
-            if (!res)
-                return;
+            var mailargs = {};
+            mailargs.to =          args.arguments.join(", ");
+            mailargs.subject =     args["-subject"];
+            mailargs.bcc =         args["-bcc"];
+            mailargs.cc =          args["-cc"];
+            mailargs.body =        args["-text"];
+            mailargs.attachments = args["-attachment"] || [];
 
-            var mailargs = new Object();
-            mailargs.subject = liberator.commands.getOption(res.opts, "-subject", undefined);
-            mailargs.bcc = liberator.commands.getOption(res.opts, "-bcc", undefined);
-            mailargs.cc = liberator.commands.getOption(res.opts, "-cc", undefined);
-            mailargs.body = liberator.commands.getOption(res.opts, "-text", undefined);
-            mailargs.attachments = liberator.commands.getOption(res.opts, "-attachment", []);
-
-            var addresses = [];
-            if (res.args)
-                addresses = addresses.concat(res.args);
+            var addresses = args.arguments;
             if (mailargs.bcc)
                 addresses = addresses.concat(mailargs.bcc);
             if (mailargs.cc)
                 addresses = addresses.concat(mailargs.cc);
 
             // TODO: is there a better way to check for validity?
-            if (addresses.some(function(recipient) { return !(/\S@\S+\.\S/.test(recipient)); }))
+            if (addresses.some(function (recipient) { return !(/\S@\S+\.\S/.test(recipient)); }))
             {
                 liberator.echoerr("Exxx: Invalid e-mail address");
                 return;
             }
 
-            if (res.args.length > 0)
-            {
-                mailargs.to = res.args.join(", ");
-            }
-
             liberator.mail.composeNewMail(mailargs);
         },
         {
-            args: [[["-subject", "-s"],    liberator.commands.OPTION_STRING],
-                   [["-attachment", "-a"], liberator.commands.OPTION_LIST],
-                   [["-bcc", "-b"],        liberator.commands.OPTION_STRING],
-                   [["-cc", "-c"],         liberator.commands.OPTION_STRING],
-                   [["-text", "-t"],       liberator.commands.OPTION_STRING]]
+            options: [[["-subject", "-s"],    liberator.commands.OPTION_STRING],
+                      [["-attachment", "-a"], liberator.commands.OPTION_LIST],
+                      [["-bcc", "-b"],        liberator.commands.OPTION_STRING],
+                      [["-cc", "-c"],         liberator.commands.OPTION_STRING],
+                      [["-text", "-t"],       liberator.commands.OPTION_STRING]]
         });
 
     liberator.commands.add(["copy[to]"],
         "Copy selected messages",
         function (args, special) { moveOrCopy(true, args); },
         {
-            completer: function (filter) { return liberator.completion.mail(filter); }
+            completer: function (filter) { return getFolderCompletions(filter); }
         });
 
     liberator.commands.add(["move[to]"],
         "Move selected messages",
         function (args, special) { moveOrCopy(false, args); },
         {
-            completer: function (filter) { return liberator.completion.mail(filter); }
+            completer: function (filter) { return getFolderCompletions(filter); }
         });
 
     liberator.commands.add(["empty[trash]"],
         "Empty trash of the current account",
-        function (args, special)
+        function ()
         {
-            if (args)
-            {
-                liberator.echoerr("E488: Trailing characters");
-                return;
-            }
-
             goDoCommand("cmd_emptyTrash");
-        });
+        },
+        { argCount: "0" });
 
     liberator.commands.add(["get[messages]"],
         "Check for new messages",
         function (args, special)
         {
-            if (args)
-            {
-                liberator.echoerr("E488: Trailing characters");
-                return;
-            }
-
             liberator.mail.getNewMessages(!special);
-        });
+        },
+        { argCount: "0" });
 
     /////////////////////////////////////////////////////////////////////////////}}}
     ////////////////////// PUBLIC SECTION //////////////////////////////////////////
@@ -732,20 +782,16 @@ liberator.Mail = function () //{{{
                     while (args.attachments.length > 0)
                     {
                         var url = args.attachments.pop();
-
-                        // Check if the file really exists
-                        var file = Components.classes["@mozilla.org/file/local;1"]
-                            .createInstance(Components.interfaces.nsILocalFile);
-                        file.initWithPath(url);
-
+                        var file = liberator.io.getFile(url);
                         if (!file.exists())
                         {
-                            liberator.echoerr("Exxx: Could attach file `" + url + "'", liberator.commandline.FORCE_SINGLELINE);
-                            continue;
+                            liberator.echoerr("Exxx: Could not attach file `" + url + "'", liberator.commandline.FORCE_SINGLELINE);
+                            return;
                         }
+
                         attachment = Components.classes["@mozilla.org/messengercompose/attachment;1"]
                             .createInstance(Components.interfaces.nsIMsgAttachment);
-                        attachment.url = "file://" + url;
+                        attachment.url = "file://" + file.path;
                         params.composeFields.addAttachment(attachment);
                     }
                 }
@@ -758,7 +804,7 @@ liberator.Mail = function () //{{{
             msgComposeService.OpenComposeWindowWithParams(null, params);
         },
 
-
+        // returns an array of nsIMsgFolder objects
         getFolders: function (filter, includeServers, includeMsgFolders)
         {
             var folders = [];
@@ -792,7 +838,16 @@ liberator.Mail = function () //{{{
             var accounts = currentAccountOnly ? [this.currentAccount]
                                               : this.getFolders("", true, false);
 
-            accounts.forEach(function (account) { account.getNewMessages(msgWindow, null); });
+            accounts.forEach(function (account) {
+                try
+                {
+                    account.getNewMessages(msgWindow, null);
+                }
+                catch (e)
+                {
+                    liberator.log("Error getting messages for account " + account.prettyName + ": " + e);
+                }
+            });
         },
 
         getStatistics: function (currentAccountOnly)
@@ -800,7 +855,7 @@ liberator.Mail = function () //{{{
             var accounts = currentAccountOnly ? [this.currentAccount]
                                               : this.getFolders("", true, false);
 
-            var unreadCount = 0, totalCount = 0, newCount = 0;;
+            var unreadCount = 0, totalCount = 0, newCount = 0;
             for (var i = 0; i < accounts.length; i++)
             {
                 var account = accounts[i];
@@ -849,14 +904,23 @@ liberator.Mail = function () //{{{
          */
         selectMessage: function (validatorFunc, canWrap, openThreads, reverse, count)
         {
+            function currentIndex()
+            {
+                var index = gDBView.selection.currentIndex;
+                if (index < 0)
+                    index = 0;
+                return index;
+            }
+
             function closedThread(index)
             {
                 if (!(gDBView.viewFlags & nsMsgViewFlagsType.kThreadedDisplay))
                     return false;
 
-                index = (typeof index == "number") ? index : gDBView.selection.currentIndex;
+                index = (typeof index == "number") ? index : currentIndex();
                 return !gDBView.isContainerOpen(index) && !gDBView.isContainerEmpty(index);
             }
+
             if (typeof validatorFunc != "function")
                 return;
 
@@ -866,7 +930,7 @@ liberator.Mail = function () //{{{
             // first try to find in current folder
             if (gDBView)
             {
-                for (var i = gDBView.selection.currentIndex + (reverse ? -1 : (openThreads && closedThread() ? 0 : 1));
+                for (var i = currentIndex() + (reverse ? -1 : (openThreads && closedThread() ? 0 : 1));
                     reverse ? (i >= 0) : (i < gDBView.rowCount);
                     reverse ? i-- : i++)
                 {
@@ -879,7 +943,7 @@ liberator.Mail = function () //{{{
                         var thread = gDBView.db.GetThreadContainingMsgHdr(msg);
                         var originalCount = count;
 
-                        for (let j = (i == gDBView.selection.currentIndex && !reverse) ? 1 : (reverse ? thread.numChildren - 1 : 0);
+                        for (let j = (i == currentIndex() && !reverse) ? 1 : (reverse ? thread.numChildren - 1 : 0);
                                  reverse ? (j >= 0) : (j < thread.numChildren);
                                  reverse ? j-- : j++)
                         {
@@ -967,8 +1031,21 @@ liberator.Mail = function () //{{{
             liberator.beep();
         },
 
-    };
+        setHTML: function (value)
+        {
+            var values = [[true,  1, gDisallow_classes_no_html],  // plaintext
+                          [false, 0, 0],                          // HTML
+                          [false, 3, gDisallow_classes_no_html]]; // sanitized/simple HTML
 
+            if (typeof(value) != "number" || value < 0 || value > 2)
+                value = 1;
+
+            gPrefBranch.setBoolPref("mailnews.display.prefer_plaintext", values[value][0]);
+            gPrefBranch.setIntPref("mailnews.display.html_as", values[value][1]);
+            gPrefBranch.setIntPref("mailnews.display.disallow_mime_handlers", values[value][2]);
+            MsgReload();
+        }
+    };
     //}}}
 }; //}}}
 
