@@ -28,7 +28,7 @@ the terms of any one of the MPL, the GPL or the LGPL.
 
 // Do NOT create instances of this class yourself, use the helper method
 // liberator.options.add() instead
-liberator.Option = function (names, description, type, defaultValue, getter, setter, validator, completer) //{{{
+liberator.Option = function (names, description, type, defaultValue, scope, getter, setter, validator, completer) //{{{
 {
     if (!names || !type)
         return null;
@@ -38,6 +38,7 @@ liberator.Option = function (names, description, type, defaultValue, getter, set
     this.name = names[0];
     this.names = names;
     this.type = type;
+    this.scope = (scope & liberator.options.OPTION_SCOPE_BOTH) || liberator.options.OPTION_SCOPE_GLOBAL; // XXX set to BOTH by default someday? - kstep
     this.description = description || "";
 
     // "", 0 are valid default values
@@ -64,23 +65,52 @@ liberator.Option = function (names, description, type, defaultValue, getter, set
         }
     }
 
-    this.__defineGetter__("value",
-        function ()
+    this.get = function(scope)
+    {
+        if (scope)
         {
-            if (this.getter)
-                this.getter.call(this);
-            return value;
-        }
-    );
-    this.__defineSetter__("value",
-        function (newValue)
+            if ((scope & this.scope) == 0) // option doesn't exist in this scope
+                return null;
+        }    
+        else
+            scope = this.scope;
+
+        var aValue;
+
+        if (liberator.has("tabs") && (scope & liberator.options.OPTION_SCOPE_LOCAL))
+            aValue = liberator.tabs.options[this.name];
+        if ((scope & liberator.options.OPTION_SCOPE_GLOBAL) && (aValue == undefined))
+            aValue = value;
+
+        if (this.getter)
+            this.getter.call(this, aValue);
+
+        return aValue;
+    };
+
+    this.set = function(newValue, scope)
+    {
+        if (scope)
         {
+            if ((scope & this.scope) == 0) // option doesn't exist in this scope
+                return null;
+        }    
+        else
+            scope = this.scope;
+
+        if (liberator.has("tabs") && (scope & liberator.options.OPTION_SCOPE_LOCAL))
+            liberator.tabs.options[this.name] = newValue;
+        if (scope & liberator.options.OPTION_SCOPE_GLOBAL)
             value = newValue;
-            this.hasChanged = true;
-            if (this.setter)
-                this.setter.call(this, value);
-        }
-    );
+
+        this.hasChanged = true;
+        if (this.setter)
+            this.setter.call(this, newValue);
+    };
+
+    this.__defineGetter__("value", this.get);
+
+    this.__defineSetter__("value", this.set);
 
     this.hasName = function (name)
     {
@@ -227,7 +257,7 @@ liberator.Options = function () //{{{
 
             var matches;
             // 1 - type, 2 - name, 3 - +-., 4 - expr
-            if (matches = args.match(/([$@&])?([\w:]+)\s*([+-.])?=\s*(.+)/))
+            if (matches = args.match(/([$@&])?([\w:]+)\s*([-+.])?=\s*(.+)/))
             {
                 if (!matches[1])
                 {
@@ -290,28 +320,50 @@ liberator.Options = function () //{{{
         });
 
     liberator.commands.add(["pref[erences]", "prefs"],
-        "Show " + liberator.config.hostApplication + " Preferences",
-        function (args, special, count, modifiers)
+        "Show " + liberator.config.hostApplication + " preferences",
+        function (args, special)
         {
-            if (!args)
+            if (special) // open Firefox settings GUI dialog
             {
-                if (special) // open firefox settings gui dialog
-                {
-                    liberator.open("about:config",
-                        (liberator.options.newtab &&
-                            (liberator.options.newtab == "all" || liberator.options.newtab.split(",").indexOf("prefs") != -1)) ?
-                                liberator.NEW_TAB : liberator.CURRENT_TAB);
-                }
-                else
-                {
-                    openPreferences();
-                }
+                liberator.open("about:config",
+                    (liberator.options.newtab &&
+                        (liberator.options.newtab == "all" || liberator.options.newtab.split(",").indexOf("prefs") != -1)) ?
+                            liberator.NEW_TAB : liberator.CURRENT_TAB);
             }
             else
             {
-                liberator.echoerr("E488: Trailing characters");
+                openPreferences();
             }
-        });
+        },
+        { argCount: "0" });
+
+    liberator.commands.add(["setl[ocal]"],
+        "Set local option",
+        function (args, special, count)
+        {
+            liberator.commands.get("set").execute(args, special, count, { scope: liberator.options.OPTION_SCOPE_LOCAL });
+        },
+        {
+            completer: function (filter, special, count)
+            {
+                return liberator.commands.get("set").completer(filter, special, count, { scope: liberator.options.OPTION_SCOPE_LOCAL });
+            }
+        }
+    );
+
+    liberator.commands.add(["setg[lobal]"],
+        "Set global option",
+        function (args, special, count)
+        {
+            liberator.commands.get("set").execute(args, special, count, { scope: liberator.options.OPTION_SCOPE_GLOBAL });
+        },
+        {
+            completer: function (filter, special, count)
+            {
+                return liberator.commands.get("set").completer(filter, special, count, { scope: liberator.options.OPTION_SCOPE_GLOBAL });
+            }
+        }
+    );
 
     // TODO: support setting multiple options at once
     liberator.commands.add(["se[t]"],
@@ -326,8 +378,9 @@ liberator.Options = function () //{{{
                     args = "all";
                     onlyNonDefault = true;
                 }
+
                 //                                1                    2       3  4       5
-                var matches = args.match(/^\s*?([a-zA-Z0-9\.\-_{}]+)([?&!])?\s*(([+-^]?)=(.*))?\s*$/);
+                var matches = args.match(/^\s*?([a-zA-Z0-9\.\-_{}]+)([?&!])?\s*(([-+^]?)=(.*))?\s*$/);
                 var name = matches[1];
                 var reset = false;
                 var invertBoolean = false;
@@ -338,7 +391,7 @@ liberator.Options = function () //{{{
                     invertBoolean = true;
 
                 if (name == "all" && reset)
-                    liberator.echoerr("You can't reset all the firefox options, it could make your browser unusable.");
+                    liberator.echoerr("You can't reset all options, it could make " + liberator.config.hostApplication + " unusable.");
                 else if (name == "all")
                     liberator.options.listPrefs(onlyNonDefault, "");
                 else if (reset)
@@ -360,9 +413,8 @@ liberator.Options = function () //{{{
                             value = false;
                             break;
                         default:
-                            var valueInt = parseInt(value, 10);
-                            if (!isNaN(valueInt))
-                                value = valueInt;
+                            if (/^\d+$/.test(value))
+                                value = parseInt(value, 10);
                     }
                     liberator.options.setPref(name, value);
                 }
@@ -381,7 +433,7 @@ liberator.Options = function () //{{{
             }
 
             //                               1        2       3       4  5       6
-            var matches = args.match(/^\s*(no|inv)?([a-z_]+)([?&!])?\s*(([+-^]?)=(.*))?\s*$/);
+            var matches = args.match(/^\s*(no|inv)?([a-z_]+)([?&!])?\s*(([-+^]?)=(.*))?\s*$/);
             if (!matches)
             {
                 liberator.echoerr("E518: Unknown option: " + args);
@@ -397,7 +449,12 @@ liberator.Options = function () //{{{
             if (name == "all")
                 all = true;
 
-            var option = liberator.options.get(name);
+            var scope = liberator.options.OPTION_SCOPE_BOTH;
+            if (modifiers && modifiers.scope)
+                scope = modifiers.scope;
+
+            var option = liberator.options.get(name, scope);
+
             if (!option && !all)
             {
                 liberator.echoerr("E518: Unknown option: " + args);
@@ -442,23 +499,23 @@ liberator.Options = function () //{{{
             {
                 if (all)
                 {
-                    liberator.options.list(onlyNonDefault);
+                    liberator.options.list(onlyNonDefault, scope);
                 }
                 else
                 {
                     if (option.type == "boolean")
-                        liberator.echo((option.value ? "  " : "no") + option.name);
+                        liberator.echo((option.get(scope) ? "  " : "no") + option.name);
                     else
-                        liberator.echo("  " + option.name + "=" + option.value);
+                        liberator.echo("  " + option.name + "=" + option.get(scope));
                 }
             }
             // write access
-            // NOTE: the behaviour is generally Vim compatible but could be
-            // improved. i.e. Vim's behaviour is pretty sloppy to no real
-            // benefit
+            // NOTE: the behavior is generally Vim compatible but could be
+            // improved. i.e. Vim's behavior is pretty sloppy to no real benefit
             else
             {
-                var currentValue = option.value;
+                var currentValue = option.get(scope);
+
                 var newValue;
 
                 switch (option.type)
@@ -471,7 +528,7 @@ liberator.Options = function () //{{{
                         }
 
                         if (invertBoolean)
-                            newValue = !option.value;
+                            newValue = !currentValue;
                         else
                             newValue = !unsetBoolean;
 
@@ -553,31 +610,120 @@ liberator.Options = function () //{{{
                 }
 
                 if (option.isValidValue(newValue))
-                    option.value = newValue;
+                {
+                    option.set(newValue, scope);
+                }
                 else
                     // FIXME: need to be able to specify more specific errors
                     liberator.echoerr("E474: Invalid argument: " + args);
             }
         },
         {
-            completer: function (filter, special) { return liberator.completion.option(filter, special); }
+            completer: function (filter, special, count, modifiers)
+            {
+                var optionCompletions = [];
+                var prefix = filter.match(/^no|inv/) || "";
+
+                if (prefix)
+                    filter = filter.replace(prefix, "");
+
+                if (special) // list completions for about:config entries
+                {
+                    var prefs = Components.classes["@mozilla.org/preferences-service;1"]
+                                .getService(Components.interfaces.nsIPrefBranch);
+                    var prefArray = prefs.getChildList("", { value: 0 });
+                    prefArray.sort();
+
+                    if (filter.length > 0 && filter.lastIndexOf("=") == filter.length - 1)
+                    {
+                        for (var i = 0; i < prefArray.length; i++)
+                        {
+                            var name = prefArray[i];
+                            if (name.match("^" + filter.substr(0, filter.length - 1) + "$" ))
+                            {
+                                var value = liberator.options.getPref(name) + "";
+                                return [filter.length + 1, [[value, ""]]];
+                            }
+                        }
+                        return [0, []];
+                    }
+
+                    for (var i = 0; i < prefArray.length; i++)
+                        optionCompletions.push([prefArray[i], liberator.options.getPref(prefArray[i])]);
+
+                    return [0, liberator.completion.filter(optionCompletions, filter)];
+                }
+
+                var scope = liberator.options.OPTION_SCOPE_BOTH;
+                if (modifiers && modifiers.scope)
+                    scope = modifiers.scope;
+
+                if (!filter)
+                {
+                    var options = [];
+
+                    for (var option in liberator.options)
+                    {
+                        if (!(option.scope & scope))
+                            continue;
+                        if (prefix && option.type != "boolean")
+                            continue;
+                        options.push([prefix + option.name, option.description]);
+                    }
+                    return [0, options];
+                }
+                // check if filter ends with =, then complete current value
+                else if (filter.length > 0 && filter.lastIndexOf("=") == filter.length - 1)
+                {
+                    filter = filter.substr(0, filter.length - 1);
+                    for (var option in liberator.options)
+                    {
+                        if (!(option.scope & scope))
+                            continue;
+                        if (option.hasName(filter))
+                        {
+                            if (option.completer)
+                                return [filter.length + 1, option.completer(filter)];
+                            return [filter.length + 1, [[option.value + "", ""]]];
+                        }
+                    }
+                    return [0, optionCompletions];
+                }
+
+                var filterLength = filter.length;
+                for (var option in liberator.options)
+                {
+                    if (!(option.scope & scope))
+                        continue;
+                    if (prefix && option.type != "boolean")
+                        continue;
+
+                    for (var j = 0; j < option.names.length; j++)
+                    {
+                        if (option.names[j].indexOf(filter) != 0)
+                            continue;
+
+                        optionCompletions.push([prefix + option.names[j], option.description]);
+                        break;
+                    }
+                }
+
+                return [0, liberator.completion.filter(optionCompletions, prefix + filter, true)];
+            }
         });
 
     liberator.commands.add(["unl[et]"],
         "Delete a variable",
         function (args, special)
         {
-            if (!args)
-            {
-                liberator.echoerr("E471: Argument required");
-                return;
-            }
+            //var names = args.split(/ /);
+            //if (typeof names == "string") names = [names];
 
-            var names = args.split(/ /);
-            if (typeof names == "string") names = [names];
-            var length = names.length;
-            for (var i = 0, name = names[i]; i < length; name = names[++i])
+            //var length = names.length;
+            //for (var i = 0, name = names[i]; i < length; name = names[++i])
+            for (var i = 0; i < args.arguments.length; i++)
             {
+                var name = args.arguments[i];
                 var reference = liberator.variableReference(name);
                 if (!reference[0])
                 {
@@ -588,7 +734,8 @@ liberator.Options = function () //{{{
 
                 delete reference[0][reference[1]];
             }
-        });
+        },
+        { argCount: "+" });
 
     /////////////////////////////////////////////////////////////////////////////}}}
     ////////////////////// PUBLIC SECTION //////////////////////////////////////////
@@ -596,6 +743,10 @@ liberator.Options = function () //{{{
 
     return {
 
+        OPTION_SCOPE_GLOBAL: 1,
+        OPTION_SCOPE_LOCAL:  2,
+        OPTION_SCOPE_BOTH:   3,
+
         __iterator__: function ()
         {
             for (var i = 0; i < options.length; i++)
@@ -609,8 +760,21 @@ liberator.Options = function () //{{{
             if (!extraInfo)
                 extraInfo = {};
 
-            var option = new liberator.Option(names, description, type, defaultValue,
-                                               extraInfo.getter, extraInfo.setter, extraInfo.validator);
+            var option = new liberator.Option(names, description, type, defaultValue, extraInfo.scope,
+                                              extraInfo.getter, extraInfo.setter, extraInfo.validator, extraInfo.completer);
+
+            if (!option)
+                return false;
+
+            for (var i = 0; i < options.length; i++)
+            {
+                if (options[i].name == option.name)
+                {
+                    // never replace for now
+                    liberator.log("Warning: '" + names[0] + "' already exists, NOT replacing existing option.", 1);
+                    return false;
+                }
+            }
 
             // quickly access options with liberator.options["wildmode"]:
             this.__defineGetter__(option.name, function () { return option.value; });
@@ -618,6 +782,7 @@ liberator.Options = function () //{{{
 
             // TODO: sort option
             options.push(option);
+            return true;
         },
 
         destroy: function ()
@@ -628,22 +793,29 @@ liberator.Options = function () //{{{
                 storePreference("dom.popup_allowed_events", popupAllowedEvents);
         },
 
-        get: function (name)
+        get: function (name, scope)
         {
+            if (!scope)
+                scope = liberator.options.OPTION_SCOPE_BOTH;
+
             for (var i = 0; i < options.length; i++)
             {
-                if (options[i].hasName(name))
+                if (options[i].hasName(name) && (options[i].scope & scope))
                     return options[i];
             }
+
             return null;
         },
 
-        list: function (onlyNonDefault)
+        list: function (onlyNonDefault, scope)
         {
             var list = ":" + liberator.util.escapeHTML(liberator.commandline.getCommand()) + "<br/>" +
                        "<table><tr align=\"left\" class=\"hl-Title\"><th>--- Options ---</th></tr>";
             var name, value, def;
 
+            if (!scope)
+                scope = liberator.options.OPTION_SCOPE_BOTH;
+
             for (var i = 0; i < options.length; i++)
             {
                 name  = options[i].name;
@@ -653,6 +825,9 @@ liberator.Options = function () //{{{
                 if (onlyNonDefault && value == def)
                     continue;
 
+                if (!(options[i].scope & scope))
+                    continue;
+
                 if (options[i].type == "boolean")
                 {
                     name = value ? "  " + name : "no" + name;
