@@ -33,19 +33,20 @@ liberator.Tabs = function () //{{{
     ////////////////////////////////////////////////////////////////////////////////
     ////////////////////// PRIVATE SECTION /////////////////////////////////////////
     /////////////////////////////////////////////////////////////////////////////{{{
+
     var tabmail;
-    var getBrowser = (function(){
+    var getBrowser = (function () {
         if (liberator.config.hostApplication == "Thunderbird")
         {
-            return function()
+            return function ()
             {
                 if (!tabmail)
                 {
                     tabmail = document.getElementById('tabmail');
-                    tabmail.__defineGetter__('mTabContainer',function(){ return this.tabContainer; });
-                    tabmail.__defineGetter__('mTabs',function(){ return this.tabContainer.childNodes; });
-                    tabmail.__defineGetter__('mCurrentTab',function(){ return this.tabContainer.selectedItem; });
-                    tabmail.__defineGetter__('mStrip',function(){ return this.tabStrip; });
+                    tabmail.__defineGetter__('mTabContainer', function () { return this.tabContainer; });
+                    tabmail.__defineGetter__('mTabs', function () { return this.tabContainer.childNodes; });
+                    tabmail.__defineGetter__('mCurrentTab', function () { return this.tabContainer.selectedItem; });
+                    tabmail.__defineGetter__('mStrip', function () { return this.tabStrip; });
                 }
                 return tabmail;
             };
@@ -126,7 +127,15 @@ liberator.Tabs = function () //{{{
                     tabs.collapsed = false;
                 }
             },
-            validator: function (value) { return (value >= 0 && value <= 2); }
+            validator: function (value) { return (value >= 0 && value <= 2); },
+            completer: function (filter)
+            {
+                return [
+                  ["0", "Never show tab bar"],
+                  ["1", "Show tab bar only if more than one tab is open"],
+                  ["2", "Always show tab bar"],
+                ];
+            },
         });
 
 
@@ -230,26 +239,7 @@ liberator.Tabs = function () //{{{
 
         liberator.mappings.add([liberator.modes.NORMAL], ["<C-^>", "<C-6>"],
             "Select the alternate tab",
-            function ()
-            {
-                if (liberator.tabs.alternate == null || liberator.tabs.getTab() == liberator.tabs.alternate)
-                {
-                    liberator.echoerr("E23: No alternate page");
-                    return;
-                }
-
-                // NOTE: this currently relies on v.tabs.index() returning the
-                // currently selected tab index when passed null
-                var index = liberator.tabs.index(liberator.tabs.alternate);
-
-                // TODO: since a tab close is more like a bdelete for us we
-                // should probably reopen the closed tab when a 'deleted'
-                // alternate is selected
-                if (index == -1)
-                    liberator.echoerr("E86: Buffer does not exist");  // TODO: This should read "Buffer N does not exist"
-                else
-                    liberator.tabs.select(index);
-            });
+            function () { liberator.tabs.selectAlternateTab(); });
     }
 
     /////////////////////////////////////////////////////////////////////////////}}}
@@ -260,9 +250,45 @@ liberator.Tabs = function () //{{{
         "Delete current buffer",
         function (args, special, count)
         {
-            liberator.tabs.remove(getBrowser().mCurrentTab, count > 0 ? count : 1, special, 0);
-        });
+            if (args)
+            {
+                args = args.toLowerCase();
+                var removed = 0;
+                var match;
+                if (match = args.match(/^(\d+):?/))
+                {
+                    liberator.tabs.remove(liberator.tabs.getTab(parseInt(match[1], 10) - 1));
+                    removed = 1;
+                }
+                else
+                {
+                    var browsers = getBrowser().browsers;
+                    for (let i = browsers.length - 1; i >= 0; i--)
+                    {
+                        var title = browsers[i].contentTitle.toLowerCase() || "";
+                        var uri = browsers[i].currentURI.spec.toLowerCase();
+                        var host = browsers[i].currentURI.host.toLowerCase();
 
+                        if (host.indexOf(args) >= 0 || uri == args ||
+                            (special && (title.indexOf(args) >= 0 || uri.indexOf(args) >= 0)))
+                        {
+                            liberator.tabs.remove(liberator.tabs.getTab(i));
+                            removed++;
+                        }
+                    }
+                }
+
+                if (removed > 0)
+                    liberator.echo(removed + " fewer tab(s)");
+                else
+                    liberator.echoerr("E94: No matching tab for " + args);
+            }
+            else // just remove the current tab
+                liberator.tabs.remove(getBrowser().mCurrentTab, count > 0 ? count : 1, special, 0);
+        },
+        { completer: function (filter) { return liberator.completion.buffer(filter); } });
+
+    // TODO: this should open in a new tab positioned directly after the current one, not at the end
     liberator.commands.add(["tab"],
         "Execute a command and tell it to output in a new tab",
         function (args)
@@ -271,37 +297,68 @@ liberator.Tabs = function () //{{{
             liberator.execute(args);
             liberator.forceNewTab = false;
         },
-        { completer: function (filter) { return liberator.completion.exTabCompletion(filter); } });
-
-    liberator.commands.add(["tabl[ast]"],
-        "Switch to the last tab",
-        function () { liberator.tabs.select("$", false); });
-
-    // TODO: count support
-    liberator.commands.add(["tabp[revious]", "tp[revious]", "tabN[ext]", "tN[ext]"],
-        "Switch to the previous tab or go [count] tabs back",
-        function (args)
         {
-            if (!args)
-                liberator.tabs.select("-1", true);
-            else if (/^\d+$/.test(args))
-                liberator.tabs.select("-" + args, true); // FIXME: urgh!
-            else
-                liberator.echoerr("E488: Trailing characters");
+            completer: function (filter) { return liberator.completion.ex(filter); }
         });
 
-    // TODO: count support
-    liberator.commands.add(["tabn[ext]", "tn[ext]"],
-        "Switch to the next or [count]th tab",
-        function (args)
+    liberator.commands.add(["tabl[ast]", "bl[ast]"],
+        "Switch to the last tab",
+        function ()
         {
-            if (!args)
+            liberator.tabs.select("$", false);
+        },
+        { argCount: "0" });
+
+    // TODO: "Zero count" if 0 specified as arg
+    liberator.commands.add(["tabp[revious]", "tp[revious]", "tabN[ext]", "tN[ext]", "bp[revious]", "bN[ext]"],
+        "Switch to the previous tab or go [count] tabs back",
+        function (args, special, count)
+        {
+            // count is ignored if an arg is specified, as per Vim
+            if (args)
             {
-                liberator.tabs.select("+1", true);
+                if (/^\d+$/.test(args))
+                    liberator.tabs.select("-" + args, true); // FIXME: urgh!
+                else
+                    liberator.echoerr("E488: Trailing characters");
             }
-            else if (/^\d+$/.test(args))
+            else if (count > 0)
             {
-                var index = parseInt(args, 10) - 1;
+                liberator.tabs.select("-" + count, true);
+            }
+            else
+            {
+                liberator.tabs.select("-1", true);
+            }
+        });
+
+    // TODO: "Zero count" if 0 specified as arg
+    liberator.commands.add(["tabn[ext]", "tn[ext]", "bn[ext]"],
+        "Switch to the next or [count]th tab",
+        function (args, special, count)
+        {
+            if (args || count > 0)
+            {
+                var index;
+
+                // count is ignored if an arg is specified, as per Vim
+                if (args)
+                {
+                    if (/^\d+$/.test(args))
+                    {
+                        index = args - 1;
+                    }
+                    else
+                    {
+                        liberator.echoerr("E488: Trailing characters");
+                        return;
+                    }
+                }
+                else
+                {
+                    index = count - 1;
+                }
+
                 if (index < liberator.tabs.count)
                     liberator.tabs.select(index, true);
                 else
@@ -309,43 +366,71 @@ liberator.Tabs = function () //{{{
             }
             else
             {
-                liberator.echoerr("E488: Trailing characters");
+                liberator.tabs.select("+1", true);
             }
         });
 
-    liberator.commands.add(["tabr[ewind]", "tabfir[st]"],
+    liberator.commands.add(["tabr[ewind]", "tabfir[st]", "br[ewind]", "bf[irst]"],
         "Switch to the first tab",
-        function () { liberator.tabs.select(0, false); });
+        function ()
+        {
+            liberator.tabs.select(0, false);
+        },
+        { argCount: "0" });
 
 
     if (liberator.config.name == "Vimperator")
     {
+        // TODO: "Zero count" if 0 specified as arg, multiple args and count ranges?
         liberator.commands.add(["b[uffer]"],
             "Switch to a buffer",
-            function (args, special) { liberator.tabs.switchTo(args, special); },
+            function (args, special, count)
+            {
+                // if a numeric arg is specified any count is ignored; if a
+                // count and non-numeric arg are both specified then E488
+                if (args && count > 0)
+                {
+                    if (/^\d+$/.test(args))
+                        liberator.tabs.switchTo(args, special);
+                    else
+                        liberator.echoerr("E488: Trailing characters");
+                }
+                else if (count > 0)
+                {
+                    liberator.tabs.switchTo(count.toString(), special);
+                }
+                else
+                {
+                    liberator.tabs.switchTo(args, special);
+                }
+            },
             { completer: function (filter) { return liberator.completion.buffer(filter); } });
 
         liberator.commands.add(["buffers", "files", "ls", "tabs"],
             "Show a list of all buffers",
             function (args, special)
             {
-                if (args)
-                {
-                    liberator.echoerr("E488: Trailing characters");
-                    return;
-                }
-
                 liberator.tabs.list(special);
-            });
+            },
+            { argCount: "0" });
 
         liberator.commands.add(["quita[ll]", "qa[ll]"],
             "Quit " + liberator.config.appName,
-            function () { liberator.quit(false); });
+            function ()
+            {
+                liberator.quit(false);
+            },
+            { argCount: "0" });
 
         liberator.commands.add(["reloada[ll]"],
             "Reload all tab pages",
-            function (args, special) { liberator.tabs.reloadAll(special); });
+            function (args, special)
+            {
+                liberator.tabs.reloadAll(special);
+            },
+            { argCount: "0" });
 
+        // TODO: add count support
         liberator.commands.add(["tabm[ove]"],
             "Move the current tab after tab N",
             function (args, special)
@@ -365,7 +450,11 @@ liberator.Tabs = function () //{{{
 
         liberator.commands.add(["tabo[nly]"],
             "Close all other tabs",
-            function () { liberator.tabs.keepOnly(getBrowser().mCurrentTab); });
+            function ()
+            {
+                liberator.tabs.keepOnly(getBrowser().mCurrentTab);
+            },
+            { argCount: "0" });
 
         liberator.commands.add(["tabopen", "t[open]", "tabnew", "tabe[dit]"],
             "Open one or more URLs in a new tab",
@@ -381,6 +470,24 @@ liberator.Tabs = function () //{{{
                     liberator.open("about:blank", where);
             },
             { completer: function (filter) { return liberator.completion.url(filter); } });
+
+        liberator.commands.add(["tabd[uplicate]"],
+            "Duplicate current tab",
+            function (args, special, count)
+            {
+                var tab = liberator.tabs.getTab();
+
+                var activate = special ? true : false;
+                if (/\btabopen\b/.test(liberator.options["activate"]))
+                    activate = !activate;
+
+                if (count < 1)
+                    count = 1;
+
+                for (var i = 0; i < count; i++)
+                    liberator.tabs.cloneTab(tab, activate);
+            },
+            { argCount: "0" });
     }
 
     if (liberator.has("session"))
@@ -449,11 +556,16 @@ liberator.Tabs = function () //{{{
                 var undoItems = eval("(" + ss.getClosedTabData(window) + ")");
                 for (var i = 0; i < undoItems.length; i++)
                     undoCloseTab(); // doesn't work with i as the index to undoCloseTab
-            });
+            },
+            { argCount: "0" });
 
         liberator.commands.add(["wqa[ll]", "wq", "xa[ll]"],
             "Save the session and quit",
-            function () { liberator.quit(true); });
+            function ()
+            {
+                liberator.quit(true);
+            },
+            { argCount: "0" });
     }
 
     /////////////////////////////////////////////////////////////////////////////}}}
@@ -466,6 +578,16 @@ liberator.Tabs = function () //{{{
 
         get count() { return getBrowser().mTabs.length; },
 
+        // used for :setlocal
+        get options()
+        {
+            var tab = this.getTab();
+            if (!tab.liberatorOptions)
+                tab.liberatorOptions = {};
+
+            return tab.liberatorOptions;
+        },
+
         get tabStrip()
         {
             if (liberator.config.hostApplication == "Firefox")
@@ -515,51 +637,35 @@ liberator.Tabs = function () //{{{
             return getBrowser().mTabContainer.selectedItem;
         },
 
-        list: function (fullmode)
+        // TODO: shouldn't that have a filter argument?
+        list: function ()
         {
-            if (fullmode)
+            // TODO: move this to liberator.tabs.get()
+            var items = liberator.completion.buffer("")[1];
+            var number, indicator, title, url;
+
+            var list = ":" + (liberator.util.escapeHTML(liberator.commandline.getCommand()) || "buffers") + "<br/>" + "<table>";
+            for (var i = 0; i < items.length; i++)
             {
-                // toggle the special buffer preview window
-                if (liberator.bufferwindow.visible())
-                {
-                    liberator.bufferwindow.hide();
-                }
+                if (i == liberator.tabs.index())
+                   indicator = " <span style=\"color: blue\">%</span> ";
+                else if (i == liberator.tabs.index(liberator.tabs.alternate))
+                   indicator = " <span style=\"color: blue\">#</span> ";
                 else
-                {
-                    var items = liberator.completion.buffer("")[1];
-                    liberator.bufferwindow.show(items);
-                    liberator.bufferwindow.selectItem(getBrowser().mTabContainer.selectedIndex);
-                }
+                   indicator = "   ";
+
+                [number, title] = items[i][0].split(/:\s+/, 2);
+                url = items[i][1];
+                url = liberator.util.escapeHTML(url);
+                title = liberator.util.escapeHTML(title);
+
+                list += "<tr><td align=\"right\">  " + number + "</td><td>" + indicator +
+                        "</td><td style=\"width: 250px; max-width: 500px; overflow: hidden;\">" + title +
+                        "</td><td><a href=\"#\" class=\"hl-URL buffer-list\">" + url + "</a></td></tr>";
             }
-            else
-            {
-                // TODO: move this to liberator.buffers.get()
-                var items = liberator.completion.buffer("")[1];
-                var number, indicator, title, url;
+            list += "</table>";
 
-                var list = ":" + (liberator.util.escapeHTML(liberator.commandline.getCommand()) || "buffers") + "<br/>" + "<table>";
-                for (var i = 0; i < items.length; i++)
-                {
-                    if (i == liberator.tabs.index())
-                       indicator = " <span style=\"color: blue\">%</span> ";
-                    else if (i == liberator.tabs.index(liberator.tabs.alternate))
-                       indicator = " <span style=\"color: blue\">#</span> ";
-                    else
-                       indicator = "   ";
-
-                    [number, title] = items[i][0].split(/:\s+/, 2);
-                    url = items[i][1];
-                    url = liberator.util.escapeHTML(url);
-                    title = liberator.util.escapeHTML(title);
-
-                    list += "<tr><td align=\"right\">  " + number + "</td><td>" + indicator +
-                            "</td><td style=\"width: 250px; max-width: 500px; overflow: hidden;\">" + title +
-                            "</td><td><a href=\"#\" class=\"hl-URL buffer-list\">" + url + "</a></td></tr>";
-                }
-                list += "</table>";
-
-                liberator.commandline.echo(list, liberator.commandline.HL_NORMAL, liberator.commandline.FORCE_MULTILINE);
-            }
+            liberator.commandline.echo(list, liberator.commandline.HL_NORMAL, liberator.commandline.FORCE_MULTILINE);
         },
 
         // wrap causes the movement to wrap around the start and end of the tab list
@@ -605,7 +711,7 @@ liberator.Tabs = function () //{{{
                             liberator.beep();
                     };
                 }
-                return function() { return null; };
+                return function () { return null; };
             })();
 
             if (typeof count != "number" || count < 1)
@@ -702,10 +808,9 @@ liberator.Tabs = function () //{{{
         switchTo: function (buffer, allowNonUnique, count, reverse)
         {
             if (buffer == "")
-            {
                 return;
-            }
-            else if (buffer != null)
+
+            if (buffer != null)
             {
                 // store this command, so it can be repeated with "B"
                 lastBufferSwitchArgs = buffer;
@@ -718,6 +823,12 @@ liberator.Tabs = function () //{{{
                     allowNonUnique = lastBufferSwitchSpecial;
             }
 
+            if (buffer == "#")
+            {
+                liberator.tabs.selectAlternateTab();
+                return;
+            }
+
             if (!count || count < 1)
                 count = 1;
             if (typeof reverse != "boolean")
@@ -766,7 +877,47 @@ liberator.Tabs = function () //{{{
             }
         },
 
-        // TODO: when restarting a session FF selects the first tab and then the
+        cloneTab: function (tab, activate)
+        {
+            var ss = Components.classes["@mozilla.org/browser/sessionstore;1"].
+                     getService(Components.interfaces.nsISessionStore);
+
+            if (!tab)
+                tab = getBrowser().mTabContainer.selectedItem;
+
+            var tabState = ss.getTabState(tab);
+            var newTab = getBrowser().addTab();
+            ss.setTabState(newTab, tabState);
+            if (activate)
+                getBrowser().mTabContainer.selectedItem = newTab;
+
+            return newTab;
+        },
+
+        selectAlternateTab: function ()
+        {
+          if (liberator.tabs.alternate == null || liberator.tabs.getTab() == liberator.tabs.alternate)
+          {
+            liberator.echoerr("E23: No alternate page");
+            return;
+          }
+
+          // NOTE: this currently relies on v.tabs.index() returning the
+          // currently selected tab index when passed null
+          var index = liberator.tabs.index(liberator.tabs.alternate);
+
+          // TODO: since a tab close is more like a bdelete for us we
+          // should probably reopen the closed tab when a 'deleted'
+          // alternate is selected
+          if (index == -1)
+            liberator.echoerr("E86: Buffer does not exist");  // TODO: This should read "Buffer N does not exist"
+          else
+            liberator.tabs.select(index);
+
+          return;
+        },
+
+        // NOTE: when restarting a session FF selects the first tab and then the
         // tab that was selected when the session was created.  As a result the
         // alternate after a restart is often incorrectly tab 1 when there
         // shouldn't be one yet.
